#
# SortingHat basic settings file.
#
# This file defines the required settings to run SortingHat.
# Due to SortingHat is a Django based app, this settings are
# based on the configuration file generated by Django by
# default.
#
# Please check the next links for details about the configuration
# in a production environment:
#
# https://docs.djangoproject.com/en/3.1/howto/deployment/checklist/
# https://docs.djangoproject.com/en/3.1/ref/settings/
#

import os

from django.core.management.utils import get_random_secret_key


BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

SILENCED_SYSTEM_CHECKS = [
    'django_mysql.E016'
]

#
# General app parameters
#

#
# You must never enable debug in production.
#
# https://docs.djangoproject.com/en/3.1/ref/settings/#std:setting-DEBUG
#

DEBUG = False

#
# ALLOWED_HOST protects the site agains CSRF attacks.
# If DEBUG is set to False, you will need to configure this parameter,
# with the host you are using to serve SortingHat.
#
# https://docs.djangoproject.com/en/3.1/ref/settings/#allowed-hosts
#

ALLOWED_HOSTS = [
    '127.0.0.1',
    'localhost',
]

#
# The secret key must be a large random value and it must be kept secret.
#
# https://docs.djangoproject.com/en/3.1/ref/settings/#secret-key
#

SECRET_KEY = get_random_secret_key()

#
# Application definition - DO NOT MODIFY
#

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'corsheaders',
    'django_rq',
    'graphene_django',
    'sortinghat.core',
]

ROOT_URLCONF = 'sortinghat.app.urls'

WSGI_APPLICATION = 'sortinghat.app.wsgi.application'

MIDDLEWARE = [
    'corsheaders.middleware.CorsMiddleware',
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

#
# Graphene - DO NOT MODIFY
#

GRAPHENE = {
    'SCHEMA': 'sortinghat.core.schema',
    'MIDDLEWARE': [
        'graphql_jwt.middleware.JSONWebTokenMiddleware',
    ],
}


#
# Authentication - DO NOT MODIFY
#

AUTHENTICATION_BACKENDS = [
    'django.contrib.auth.backends.ModelBackend',
    'graphql_jwt.backends.JSONWebTokenBackend',
]

#
# Password validation
#
# https://docs.djangoproject.com/en/3.1/ref/settings/#auth-password-validators
#

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]

#
# Cross-Origin Resource Sharing (CORS)
#
# You'll HAVE TO configure the origins that are authorized to make
# cross-site HTTP requests. Check the following link to understand
# the possibilities and parameters you can use.
#
# https://github.com/adamchainz/django-cors-headers#configuration
#

CORS_ALLOWED_ORIGINS = [
    'http://localhost:8080',
]

CORS_ALLOW_CREDENTIALS = True

#
# Static files (CSS, JavaScript, Images)
#
# https://docs.djangoproject.com/en/3.2/howto/static-files/
#

STATIC_URL = '/'

# UI static files will be copied to the next path when
# 'collectstatic' is run.
# If you are serving these files in a dedicated server, you will
# need to copy them to their final destination.

STATIC_ROOT = "/tmp/sortinghat/static/"

# MEDIA_URL is only needed when DEBUG is set to True.
# Modify this URL if you want to run the server in developer mode.

MEDIA_URL = 'http://media.localhost/'

#
# Internationalization
#
# https://docs.djangoproject.com/en/3.1/topics/i18n/
#
#

LANGUAGE_CODE = 'en-us'
USE_I18N = True
USE_L10N = True

#
# Time Zone
#

USE_TZ = True
TIME_ZONE = 'UTC'

#
# SortingHat Logging
#
# https://docs.djangoproject.com/en/3.1/topics/logging/#configuring-logging
#

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'simple': {
            'format': '[{asctime}] {message}',
            'style': '{',
        },
        'verbose': {
            'format': '[{asctime} - {levelname} - {name}:{lineno}] {message}',
            'style': '{',
        },
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'formatter': 'verbose',
        },
    },
    'root': {
        'handlers': ['console'],
        'level': 'INFO',
    },
}

#
# SortingHat database
#
# You MUST set the database parameters in order to run
# SortingHat.
#

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.mysql',
        'HOST': '127.0.0.1',
        'PORT': 3306,
        'USER': 'root',
        'PASSWORD': 'root',
        'NAME': 'test_sh',
        'OPTIONS': {'charset': 'utf8mb4'},
    }
}

#
# SortingHat workers
#
# SortingHat uses RQ to run background and async jobs.
# You'll HAVE TO set the next parameters in order to run
# them in the background. If not, set ASYNC to 'False'.
#
# Take into account RQ uses Redis database. You have more
# info about these parameters on the following link:
#
# https://github.com/rq/django-rq
#

RQ_QUEUES = {
    'default': {
        'HOST': 'localhost',
        'PORT': 6379,
        'ASYNC': True,
        'DB': 0
    }
}

#
# SortingHat Core parameters
#

# Require authentication when using the API.
# You shouldn't deactivate this option unless you are debugging
# the system or running it in a trusted and safe environment.

SORTINGHAT_AUTHENTICATION_REQUIRED = False

#
# API default page size
#

SORTINGHAT_API_PAGE_SIZE = 10

#
# genderize.io token, used only for gender recommendations
#

SORTINGHAT_GENDERIZE_API_KEY = None



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

import logging


from .enrich import Enrich, metadata
from ..elastic_analyzer import Analyzer as BaseAnalyzer
from ..elastic_mapping import Mapping as BaseMapping


logger = logging.getLogger(__name__)

SLACK_URL = "https://slack.com/"


class Analyzer(BaseAnalyzer):
    @staticmethod
    def get_elastic_analyzers(es_major):
        settings = """
            {
                "settings": {
                    "analysis": {
                        "analyzer": {
                            "my_stop_analyzer": {
                                "tokenizer": "lowercase",
                                "filter": [
                                    "english_stop",
                                    "ad_hoc_stop"
                                ]
                            }
                        },
                        "filter": {
                            "ad_hoc_stop": {
                                "type": "stop",
                                "stopwords": [
                                    "'ll","'tis","'twas","'ve","10","39","a","a's","able","ableabout","about","above",
                                    "abroad","abst","accordance","according","accordingly","across","act","actually","ad",
                                    "added","adj","adopted","ae","af","affected","affecting","affects","after","afterwards",
                                    "ag","again","against","ago","ah","ahead","ai","ain't","aint","al","all","allow",
                                    "allows","almost","alone","along","alongside","already","also","although","always","am",
                                    "amid","amidst","among","amongst","amoungst","amount","an","and","announce","another",
                                    "any","anybody","anyhow","anymore","anyone","anything","anyway","anyways","anywhere",
                                    "ao","apart","apparently","appear","appreciate","appropriate","approximately","aq","ar",
                                    "are","area","areas","aren","aren't","arent","arise","around","arpa","as","aside","ask",
                                    "asked","asking","asks","associated","at","au","auth","available","aw","away","awfully",
                                    "az","b","ba","back","backed","backing","backs","backward","backwards","bb","bd","be",
                                    "became","because","become","becomes","becoming","been","before","beforehand","began",
                                    "begin","beginning","beginnings","begins","behind","being","beings","believe","below",
                                    "beside","besides","best","better","between","beyond","bf","bg","bh","bi","big","bill",
                                    "billion","biol","bj","bm","bn","bo","both","bottom","br","brief","briefly","bs","bt",
                                    "but","buy","bv","bw","by","bz","c","c'mon","c's","ca","call","came","can","can't",
                                    "cannot","cant","caption","case","cases","cause","causes","cc","cd","certain",
                                    "certainly","cf","cg","ch","changes","ci","ck","cl","clear","clearly","click","cm",
                                    "cmon","cn","co","co.","com","come","comes","computer","con","concerning",
                                    "consequently","consider","considering","contain","containing","contains","copy",
                                    "corresponding","could","could've","couldn","couldn't","couldnt","course","cr","cry",
                                    "cs","cu","currently","cv","cx","cy","cz","d","dare","daren't","darent","date","de",
                                    "dear","definitely","describe","described","despite","detail","did","didn","didn't",
                                    "didnt","differ","different","differently","directly","dj","dk","dm","do","does",
                                    "doesn","doesn't","doesnt","doing","don","don't","done","dont","doubtful","down",
                                    "downed","downing","downs","downwards","due","during","dz","e","each","early","ec","ed",
                                    "edu","ee","effect","eg","eh","eight","eighty","either","eleven","else","elsewhere",
                                    "empty","end","ended","ending","ends","enough","entirely","er","es","especially","et",
                                    "et-al","etc","even","evenly","ever","evermore","every","everybody","everyone",
                                    "everything","everywhere","ex","exactly","example","except","f","face","faces","fact",
                                    "facts","fairly","far","farther","felt","few","fewer","ff","fi","fifteen","fifth",
                                    "fifty","fify","fill","find","finds","fire","first","five","fix","fj","fk","fm","fo",
                                    "followed","following","follows","for","forever","former","formerly","forth","forty",
                                    "forward","found","four","fr","free","from","front","full","fully","further",
                                    "furthered","furthering","furthermore","furthers","fx","g","ga","gave","gb","gd","ge",
                                    "general","generally","get","gets","getting","gf","gg","gh","gi","give","given","gives",
                                    "giving","gl","gm","gmt","gn","go","goes","going","gone","good","goods","got","gotten",
                                    "gov","gp","gq","gr","great","greater","greatest","greetings","group","grouped",
                                    "grouping","groups","gs","gt","gu","gw","gy","h","had","hadn't","hadnt","half",
                                    "happens","hardly","has","hasn","hasn't","hasnt","have","haven","haven't","havent",
                                    "having","he","he'd","he'll","he's","hed","hell","hello","help","hence","her","here",
                                    "here's","hereafter","hereby","herein","heres","hereupon","hers","herself",
                                    "herse","hes","hi","hid","high","higher","highest","him","himself","himse",
                                    "his","hither","hk","hm","hn","home","homepage","hopefully","how","how'd","how'll",
                                    "how's","howbeit","however","hr","ht","htm","html","http","hu","hundred","i","i'd",
                                    "i'll","i'm","i've","i.e.","id","ie","if","ignored","ii","il","ill","im","immediate",
                                    "immediately","importance","important","in","inasmuch","inc","inc.","indeed","index",
                                    "indicate","indicated","indicates","information","inner","inside","insofar","instead",
                                    "int","interest","interested","interesting","interests","into","invention","inward",
                                    "io","iq","ir","is","isn","isn't","isnt","it","it'd","it'll","it's","itd","itll","its",
                                    "itself","itse","ive","j","je","jm","jo","join","jp","just","k","ke","keep",
                                    "keeps","kept","keys","kg","kh","ki","kind","km","kn","knew","know","known","knows",
                                    "kp","kr","kw","ky","kz","l","la","large","largely","last","lately","later","latest",
                                    "latter","latterly","lb","lc","least","length","less","lest","let","let's","lets","li",
                                    "like","liked","likely","likewise","line","little","lk","ll","long","longer","longest",
                                    "look","looking","looks","low","lower","lr","ls","lt","ltd","lu","lv","ly","m","ma",
                                    "made","mainly","make","makes","making","man","many","may","maybe","mayn't","maynt",
                                    "mc","md","me","mean","means","meantime","meanwhile","member","members","men","merely",
                                    "mg","mh","microsoft","might","might've","mightn't","mightnt","mil","mill","million",
                                    "mine","minus","miss","mk","ml","mm","mn","mo","more","moreover","most","mostly","move",
                                    "mp","mq","mr","mrs","ms","msie","mt","mu","much","mug","must","must've","mustn't",
                                    "mustnt","mv","mw","mx","my","myself","myse","mz","n","na","name","namely","nay",
                                    "nc","nd","ne","near","nearly","necessarily","necessary","need","needed","needing",
                                    "needn't","neednt","needs","neither","net","netscape","never","neverf","neverless",
                                    "nevertheless","new","newer","newest","next","nf","ng","ni","nine","ninety","nl","no",
                                    "no-one","nobody","non","none","nonetheless","noone","nor","normally","nos","not",
                                    "noted","nothing","notwithstanding","novel","now","nowhere","np","nr","nu","null",
                                    "number","numbers","nz","o","obtain","obtained","obviously","of","off","often","oh",
                                    "ok","okay","old","older","oldest","om","omitted","on","once","one","one's","ones",
                                    "only","onto","open","opened","opening","opens","opposite","or","ord","order","ordered",
                                    "ordering","orders","org","other","others","otherwise","ought","oughtn't","oughtnt",
                                    "our","ours","ourselves","out","outside","over","overall","owing","own","p","pa","page",
                                    "pages","part","parted","particular","particularly","parting","parts","past","pe","per",
                                    "perhaps","pf","pg","ph","pk","pl","place","placed","places","please","plus","pm",
                                    "pmid","pn","point","pointed","pointing","points","poorly","possible","possibly",
                                    "potentially","pp","pr","predominantly","present","presented","presenting","presents",
                                    "presumably","previously","primarily","probably","problem","problems","promptly",
                                    "proud","provided","provides","pt","put","puts","pw","py","q","qa","que","quickly",
                                    "quite","qv","r","ran","rather","rd","re","readily","really","reasonably","recent",
                                    "recently","ref","refs","regarding","regardless","regards","related","relatively",
                                    "research","reserved","respectively","resulted","resulting","results","right","ring",
                                    "ro","room","rooms","round","ru","run","rw","s","sa","said","same","saw","say","saying",
                                    "says","sb","sc","sd","se","sec","second","secondly","seconds","section","see","seeing",
                                    "seem","seemed","seeming","seems","seen","sees","self","selves","sensible","sent",
                                    "serious","seriously","seven","seventy","several","sg","sh","shall","shan't","shant",
                                    "she","she'd","she'll","she's","shed","shell","shes","should","should've","shouldn",
                                    "shouldn't","shouldnt","show","showed","showing","shown","showns","shows","si","side",
                                    "sides","significant","significantly","similar","similarly","since","sincere","site",
                                    "six","sixty","sj","sk","sl","slightly","sm","small","smaller","smallest","sn","so",
                                    "some","somebody","someday","somehow","someone","somethan","something","sometime",
                                    "sometimes","somewhat","somewhere","soon","sorry","specifically","specified","specify",
                                    "specifying","sr","st","state","states","still","stop","strongly","su","sub",
                                    "substantially","successfully","such","sufficiently","suggest","sup","sure","sv","sy",
                                    "system","sz","t","t's","take","taken","taking","tc","td","tell","ten","tends","test",
                                    "text","tf","tg","th","than","thank","thanks","thanx","that","that'll","that's",
                                    "that've","thatll","thats","thatve","the","their","theirs","them","themselves","then",
                                    "thence","there","there'd","there'll","there're","there's","there've","thereafter",
                                    "thereby","thered","therefore","therein","therell","thereof","therere","theres",
                                    "thereto","thereupon","thereve","these","they","they'd","they'll","they're","they've",
                                    "theyd","theyll","theyre","theyve","thick","thin","thing","things","think","thinks",
                                    "third","thirty","this","thorough","thoroughly","those","thou","though","thoughh",
                                    "thought","thoughts","thousand","three","throug","through","throughout","thru","thus",
                                    "til","till","tip","tis","tj","tk","tm","tn","to","today","together","too","took","top",
                                    "toward","towards","tp","tr","tried","tries","trillion","truly","try","trying","ts",
                                    "tt","turn","turned","turning","turns","tv","tw","twas","twelve","twenty","twice","two",
                                    "tz","u","ua","ug","uk","um","un","under","underneath","undoing","unfortunately",
                                    "unless","unlike","unlikely","until","unto","up","upon","ups","upwards","us","use",
                                    "used","useful","usefully","usefulness","uses","using","usually","uucp","uy","uz","v",
                                    "va","value","various","vc","ve","versus","very","vg","vi","via","viz","vn","vol",
                                    "vols","vs","vu","w","want","wanted","wanting","wants","was","wasn","wasn't","wasnt",
                                    "way","ways","we","we'd","we'll","we're","we've","web","webpage","website","wed",
                                    "welcome","well","wells","went","were","weren","weren't","werent","weve","wf","what",
                                    "what'd","what'll","what's","what've","whatever","whatll","whats","whatve","when",
                                    "when'd","when'll","when's","whence","whenever","where","where'd","where'll","where's",
                                    "whereafter","whereas","whereby","wherein","wheres","whereupon","wherever","whether",
                                    "which","whichever","while","whilst","whim","whither","who","who'd","who'll","who's",
                                    "whod","whoever","whole","wholl","whom","whomever","whos","whose","why","why'd",
                                    "why'll","why's","widely","width","will","willing","wish","with","within","without",
                                    "won","won't","wonder","wont","words","work","worked","working","works","world","would",
                                    "would've","wouldn","wouldn't","wouldnt","ws","www","x","y","ye","year","years","yes",
                                    "yet","you","you'd","you'll","you're","you've","youd","youll","young","younger",
                                    "youngest","your","youre","yours","yourself","yourselves","youve","yt","yu","z","za",
                                    "zero","zm","zr"]
                            },
                            "english_stop": {
                                "type":       "stop",
                                "stopwords":  "_english_"
                            }
                        }
                    }
                }
            } """
        return {"items": settings}


class Mapping(BaseMapping):

    @staticmethod
    def get_elastic_mappings(es_major):
        """Get Elasticsearch mapping.

        :param es_major: major version of Elasticsearch, as string
        :returns:        dictionary with a key, 'items', with the mapping
        """

        mapping = """
        {
            "properties": {
                "text_analyzed": {
                  "type": "text",
                  "fielddata": true,
                  "analyzer": "my_stop_analyzer"
                }
           }
        } """

        return {"items": mapping}


class SlackEnrich(Enrich):

    analyzer = Analyzer
    mapping = Mapping

    def get_field_author(self):
        return "user_data"

    def get_sh_identity(self, item, identity_field=None):
        identity = {
            'username': None,
            'name': None,
            'email': None
        }

        from_ = item
        if isinstance(item, dict) and 'data' in item:
            if self.get_field_author() not in item['data']:
                # Message from bot. For the rare cases where both user
                # and bot_id are not present, an empty identity is returned
                identity['username'] = item['data'].get('bot_id', None)
                return identity
            from_ = item['data'][self.get_field_author()]

        if not from_:
            return identity
        identity['username'] = from_['name']
        identity['name'] = from_['name']
        if 'real_name' in from_:
            identity['name'] = from_['real_name']
        if 'profile' in from_:
            if 'email' in from_['profile']:
                identity['email'] = from_['profile']['email']
        return identity

    def get_identities(self, item):
        """ Return the identities from an item """

        identity = self.get_sh_identity(item)
        yield identity

    def get_project_repository(self, eitem):
        repo = eitem['origin']
        repo = repo.replace(SLACK_URL, "")  # only the channel id is included for the mapping
        return repo

    @metadata
    def get_rich_item(self, item):
        eitem = {}

        self.copy_raw_fields(self.RAW_FIELDS_COPY, item, eitem)

        # The real data
        message = item['data']

        eitem["reply_count"] = 0  # be sure it is always included

        # data fields to copy
        copy_fields = ["text", "type", "reply_count", "subscribed", "subtype",
                       "unread_count", "user"]
        for f in copy_fields:
            if f in message:
                eitem[f] = message[f]
            else:
                eitem[f] = None

        eitem['text_analyzed'] = eitem['text']

        eitem['number_attachs'] = 0
        if 'attachments' in message and message['attachments']:
            eitem['number_attachs'] = len(message['attachments'])

        eitem['reaction_count'] = 0
        if 'reactions' in message:
            eitem['reaction_count'] = len(message['reactions'])
            eitem['reactions'] = []
            for rdata in message['reactions']:
                # {
                #         "count": 2,
                #         "users": [
                #            "U38J51N7J",
                #            "U3Q0VLHU3"
                #         ],
                #         "name": "+1"
                # }
                for i in range(0, rdata['count']):
                    eitem['reactions'].append(rdata["name"])

        if 'files' in message:
            eitem['number_files'] = len(message['files'])
            message_file_size = 0
            for file in message['files']:
                message_file_size += file.get('size', 0)
            eitem['message_file_size'] = message_file_size

        if 'user_data' in message and message['user_data']:
            eitem['team_id'] = message['user_data']['team_id']
            if 'tz_offset' in message['user_data']:
                eitem['tz'] = message['user_data']['tz_offset']
                # tz must be in -12h to 12h interval, so seconds -> hours
                eitem['tz'] = round(int(eitem['tz']) / (60 * 60))
            if 'is_admin' in message['user_data']:
                eitem['is_admin'] = message['user_data']['is_admin']
            if 'is_owner' in message['user_data']:
                eitem['is_owner'] = message['user_data']['is_owner']
            if 'is_primary_owner' in message['user_data']:
                eitem['is_primary_owner'] = message['user_data']['is_primary_owner']
            if 'profile' in message['user_data']:
                if 'title' in message['user_data']['profile']:
                    eitem['profile_title'] = message['user_data']['profile']['title']
                eitem['avatar'] = message['user_data']['profile']['image_32']

        # Channel info
        channel = message['channel_info']
        eitem['channel_name'] = channel['name']
        eitem['channel_id'] = channel['id']
        eitem['channel_created'] = channel['created']
        # Due to a Slack API change, the list of members is returned paginated, thus the new attribute `num_members`
        # has been added to the Slack Perceval backend. In order to avoid breaking changes, the former
        # variable `members` is kept and used only if `num_members` is not present in the input item.
        eitem['channel_member_count'] = channel['num_members'] if 'num_members' in channel else len(channel['members'])
        if 'topic' in channel:
            eitem['channel_topic'] = channel['topic']
        if 'purpose' in channel:
            eitem['channel_purpose'] = channel['purpose']
        channel_bool_fields = ['is_archived', 'is_general', 'is_starred']
        for field in channel_bool_fields:
            eitem['channel_' + field] = 0
            if field in channel and channel[field]:
                eitem['channel_' + field] = 1

        eitem = self.__convert_booleans(eitem)

        if self.sortinghat:
            eitem.update(self.get_item_sh(item))

        if self.prjs_map:
            eitem.update(self.get_item_project(eitem))

        eitem.update(self.get_grimoire_fields(item["metadata__updated_on"], "message"))

        self.add_repository_labels(eitem)
        self.add_metadata_filter_raw(eitem)
        return eitem

    def __convert_booleans(self, eitem):
        """ Convert True/False to 1/0 for better kibana processing """

        for field in eitem.keys():
            if isinstance(eitem[field], bool):
                if eitem[field]:
                    eitem[field] = 1
                else:
                    eitem[field] = 0
        return eitem  # not needed becasue we are modifying directly the dict



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

from .elastic import ElasticOcean
from ..elastic_mapping import Mapping as BaseMapping
from ..identities.git import GitIdentities
from ..enriched.utils import anonymize_url


class Mapping(BaseMapping):

    @staticmethod
    def get_elastic_mappings(es_major):
        """Get Elasticsearch mapping.

        Ensure data.message is string, since it can be very large

        :param es_major: major version of Elasticsearch, as string
        :returns:        dictionary with a key, 'items', with the mapping
        """

        mapping = '''
         {
            "dynamic":true,
            "properties": {
                "data": {
                    "properties": {
                        "message": {
                            "type": "text",
                            "index": true
                        },
                        "AuthorDate": {
                            "type": "date",
                            "format":
                                "EEE MMM d HH:mm:ss yyyy Z||EEE MMM d HH:mm:ss yyyy||strict_date_optional_time||epoch_millis"
                        },
                        "CommitDate": {
                            "type": "date",
                            "format":
                                "EEE MMM d HH:mm:ss yyyy Z||EEE MMM d HH:mm:ss yyyy||strict_date_optional_time||epoch_millis"
                        }
                    }
                }
            }
        }
        '''

        return {"items": mapping}


class GitOcean(ElasticOcean):
    """Git Ocean feeder"""

    mapping = Mapping
    identities = GitIdentities

    def _fix_item(self, item):
        item['origin'] = anonymize_url(item['origin'])
        item['tag'] = anonymize_url(item['tag'])

    @classmethod
    def get_perceval_params_from_url(cls, url):
        params = []
        tokens = url.split(' ', 1)  # Just split the URL not the filter
        url = tokens[0]
        params.append(url)

        return params



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#   Quan Zhou <quan@bitergia.com>
#

import argparse
import logging
import sys

import requests

import pkg_resources

from grimoire_elk.errors import ElasticError
from grimoire_elk.elastic import ElasticSearch
# Connectors for Graal
from graal.backends.core.coqua import CoQua, CoQuaCommand
from graal.backends.core.cocom import CoCom, CoComCommand
from graal.backends.core.codep import CoDep, CoDepCommand
from graal.backends.core.colic import CoLic, CoLicCommand
# Connectors for Perceval
from grimoire_elk.raw.hyperkitty import HyperKittyOcean
from perceval.backends.core.askbot import Askbot, AskbotCommand
from perceval.backends.core.bugzilla import Bugzilla, BugzillaCommand
from perceval.backends.core.bugzillarest import BugzillaREST, BugzillaRESTCommand
from perceval.backends.core.confluence import Confluence, ConfluenceCommand
from perceval.backends.core.discourse import Discourse, DiscourseCommand
from perceval.backends.core.dockerhub import DockerHub, DockerHubCommand
from perceval.backends.core.gerrit import Gerrit, GerritCommand
from perceval.backends.core.git import Git, GitCommand
from perceval.backends.core.github import GitHub, GitHubCommand
from perceval.backends.core.githubql import GitHubQL, GitHubQLCommand
from perceval.backends.core.gitlab import GitLab, GitLabCommand
from perceval.backends.core.gitter import Gitter, GitterCommand
from perceval.backends.core.googlehits import GoogleHits, GoogleHitsCommand
from perceval.backends.core.groupsio import Groupsio, GroupsioCommand
from perceval.backends.core.hyperkitty import HyperKitty, HyperKittyCommand
from perceval.backends.core.jenkins import Jenkins, JenkinsCommand
from perceval.backends.core.jira import Jira, JiraCommand
from perceval.backends.core.launchpad import Launchpad, LaunchpadCommand
from perceval.backends.core.mattermost import Mattermost, MattermostCommand
from perceval.backends.core.mbox import MBox, MBoxCommand
from perceval.backends.core.mediawiki import MediaWiki, MediaWikiCommand
from perceval.backends.core.meetup import Meetup, MeetupCommand
from perceval.backends.core.nntp import NNTP, NNTPCommand
from perceval.backends.core.pagure import Pagure, PagureCommand
from perceval.backends.core.phabricator import Phabricator, PhabricatorCommand
from perceval.backends.core.pipermail import Pipermail, PipermailCommand
from perceval.backends.core.twitter import Twitter, TwitterCommand
from perceval.backends.puppet.puppetforge import PuppetForge, PuppetForgeCommand
from perceval.backends.core.redmine import Redmine, RedmineCommand
from perceval.backends.core.rocketchat import RocketChat, RocketChatCommand
from perceval.backends.core.rss import RSS, RSSCommand
from perceval.backends.core.slack import Slack, SlackCommand
from perceval.backends.core.stackexchange import StackExchange, StackExchangeCommand
from perceval.backends.core.supybot import Supybot, SupybotCommand
from perceval.backends.core.telegram import Telegram, TelegramCommand
from perceval.backends.mozilla.crates import Crates, CratesCommand
from perceval.backends.mozilla.kitsune import Kitsune, KitsuneCommand
from perceval.backends.mozilla.mozillaclub import MozillaClub, MozillaClubCommand
from perceval.backends.mozilla.remo import ReMo, ReMoCommand
from perceval.backends.opnfv.functest import Functest, FunctestCommand
from perceval.backends.weblate.weblate import Weblate, WeblateCommand

# Connectors for EnrichOcean
from .enriched.askbot import AskbotEnrich
from .enriched.bugzilla import BugzillaEnrich
from .enriched.bugzillarest import BugzillaRESTEnrich
from .enriched.cocom import CocomEnrich
from .enriched.colic import ColicEnrich
from .enriched.dockerdeps import Dockerdeps
from .enriched.dockersmells import Dockersmells
from .enriched.confluence import ConfluenceEnrich
from .enriched.crates import CratesEnrich
from .enriched.discourse import DiscourseEnrich
from .enriched.dockerhub import DockerHubEnrich
from .enriched.functest import FunctestEnrich
from .enriched.gerrit import GerritEnrich
from .enriched.git import GitEnrich
from .enriched.github import GitHubEnrich
from .enriched.githubql import GitHubQLEnrich
from .enriched.github2 import GitHubEnrich2
from .enriched.gitlab import GitLabEnrich
from .enriched.gitter import GitterEnrich
from .enriched.google_hits import GoogleHitsEnrich
from .enriched.groupsio import GroupsioEnrich
from .enriched.hyperkitty import HyperKittyEnrich
from .enriched.jenkins import JenkinsEnrich
from .enriched.jira import JiraEnrich
from .enriched.kitsune import KitsuneEnrich
from .enriched.launchpad import LaunchpadEnrich
from .enriched.mattermost import MattermostEnrich
from .enriched.mbox import MBoxEnrich
from .enriched.mediawiki import MediaWikiEnrich
from .enriched.meetup import MeetupEnrich
from .enriched.mozillaclub import MozillaClubEnrich
from .enriched.nntp import NNTPEnrich
from .enriched.pagure import PagureEnrich
from .enriched.phabricator import PhabricatorEnrich
from .enriched.pipermail import PipermailEnrich
from .enriched.puppetforge import PuppetForgeEnrich
from .enriched.redmine import RedmineEnrich
from .enriched.remo import ReMoEnrich
from .enriched.rocketchat import RocketChatEnrich
from .enriched.rss import RSSEnrich
from .enriched.slack import SlackEnrich
from .enriched.stackexchange import StackExchangeEnrich
from .enriched.supybot import SupybotEnrich
from .enriched.telegram import TelegramEnrich
from .enriched.twitter import TwitterEnrich
from .enriched.weblate import WeblateEnrich
# Connectors for Ocean
from .raw.askbot import AskbotOcean
from .raw.bugzilla import BugzillaOcean
from .raw.bugzillarest import BugzillaRESTOcean
from .raw.confluence import ConfluenceOcean
from .raw.crates import CratesOcean
from .raw.discourse import DiscourseOcean
from .raw.dockerhub import DockerHubOcean
from .raw.elastic import ElasticOcean
from .raw.functest import FunctestOcean
from .raw.gerrit import GerritOcean
from .raw.git import GitOcean
from .raw.github import GitHubOcean
from .raw.githubql import GitHubQLOcean
from .raw.gitlab import GitLabOcean
from .raw.gitter import GitterOcean
from .raw.google_hits import GoogleHitsOcean
from .raw.graal import GraalOcean
from .raw.groupsio import GroupsioOcean
from .raw.jenkins import JenkinsOcean
from .raw.jira import JiraOcean
from .raw.kitsune import KitsuneOcean
from .raw.launchpad import LaunchpadOcean
from .raw.mattermost import MattermostOcean
from .raw.mbox import MBoxOcean
from .raw.mediawiki import MediaWikiOcean
from .raw.meetup import MeetupOcean
from .raw.mozillaclub import MozillaClubOcean
from .raw.nntp import NNTPOcean
from .raw.pagure import PagureOcean
from .raw.phabricator import PhabricatorOcean
from .raw.pipermail import PipermailOcean
from .raw.puppetforge import PuppetForgeOcean
from .raw.redmine import RedmineOcean
from .raw.remo import ReMoOcean
from .raw.rocketchat import RocketChatOcean
from .raw.rss import RSSOcean
from .raw.slack import SlackOcean
from .raw.stackexchange import StackExchangeOcean
from .raw.supybot import SupybotOcean
from .raw.telegram import TelegramOcean
from .raw.twitter import TwitterOcean
from .raw.weblate import WeblateOcean

logger = logging.getLogger(__name__)

kibiter_version = None

ENTRY_POINT_NAME = "grimoire_elk"


def get_connector_from_name(name):

    # Remove extra data from data source section: remo:activities
    name = name.split(":")[0]
    found = None
    connectors = get_connectors()

    for cname in connectors:
        if cname == name:
            found = connectors[cname]

    return found


def get_connector_name(cls):
    found = None
    connectors = get_connectors()

    for cname in connectors:
        for con in connectors[cname]:
            if cls == con:
                if found:
                    # The canonical name is included in the classname
                    if cname in cls.__name__.lower():
                        found = cname
                else:
                    found = cname
    return found


def get_connector_name_from_cls_name(cls_name):
    found = None
    connectors = get_connectors()

    for cname in connectors:
        for con in connectors[cname]:
            if not con:
                continue
            if cls_name == con.__name__:
                if found:
                    # The canonical name is included in the classname
                    if cname in con.__name__.lower():
                        found = cname
                else:
                    found = cname
    return found


def get_connectors():

    connectors = {"askbot": [Askbot, AskbotOcean, AskbotEnrich, AskbotCommand],
                  "bugzilla": [Bugzilla, BugzillaOcean, BugzillaEnrich, BugzillaCommand],
                  "bugzillarest": [BugzillaREST, BugzillaRESTOcean, BugzillaRESTEnrich, BugzillaRESTCommand],
                  "cocom": [CoCom, GraalOcean, CocomEnrich, CoComCommand],
                  "colic": [CoLic, GraalOcean, ColicEnrich, CoLicCommand],
                  "dockerdeps": [CoDep, GraalOcean, Dockerdeps, CoDepCommand],
                  "dockersmells": [CoQua, GraalOcean, Dockersmells, CoQuaCommand],
                  "confluence": [Confluence, ConfluenceOcean, ConfluenceEnrich, ConfluenceCommand],
                  "crates": [Crates, CratesOcean, CratesEnrich, CratesCommand],
                  "discourse": [Discourse, DiscourseOcean, DiscourseEnrich, DiscourseCommand],
                  "dockerhub": [DockerHub, DockerHubOcean, DockerHubEnrich, DockerHubCommand],
                  "functest": [Functest, FunctestOcean, FunctestEnrich, FunctestCommand],
                  "gerrit": [Gerrit, GerritOcean, GerritEnrich, GerritCommand],
                  "git": [Git, GitOcean, GitEnrich, GitCommand],
                  "github": [GitHub, GitHubOcean, GitHubEnrich, GitHubCommand],
                  "githubql": [GitHubQL, GitHubQLOcean, GitHubQLEnrich, GitHubQLCommand],
                  "github2": [GitHub, GitHubOcean, GitHubEnrich2, GitHubCommand],
                  "gitlab": [GitLab, GitLabOcean, GitLabEnrich, GitLabCommand],
                  "gitter": [Gitter, GitterOcean, GitterEnrich, GitterCommand],
                  "google_hits": [GoogleHits, GoogleHitsOcean, GoogleHitsEnrich, GoogleHitsCommand],
                  "groupsio": [Groupsio, GroupsioOcean, GroupsioEnrich, GroupsioCommand],
                  "hyperkitty": [HyperKitty, HyperKittyOcean, HyperKittyEnrich, HyperKittyCommand],
                  "jenkins": [Jenkins, JenkinsOcean, JenkinsEnrich, JenkinsCommand],
                  "jira": [Jira, JiraOcean, JiraEnrich, JiraCommand],
                  "kitsune": [Kitsune, KitsuneOcean, KitsuneEnrich, KitsuneCommand],
                  "launchpad": [Launchpad, LaunchpadOcean, LaunchpadEnrich, LaunchpadCommand],
                  "mattermost": [Mattermost, MattermostOcean, MattermostEnrich, MattermostCommand],
                  "mbox": [MBox, MBoxOcean, MBoxEnrich, MBoxCommand],
                  "mediawiki": [MediaWiki, MediaWikiOcean, MediaWikiEnrich, MediaWikiCommand],
                  "meetup": [Meetup, MeetupOcean, MeetupEnrich, MeetupCommand],
                  "mozillaclub": [MozillaClub, MozillaClubOcean, MozillaClubEnrich, MozillaClubCommand],
                  "nntp": [NNTP, NNTPOcean, NNTPEnrich, NNTPCommand],
                  "pagure": [Pagure, PagureOcean, PagureEnrich, PagureCommand],
                  "phabricator": [Phabricator, PhabricatorOcean, PhabricatorEnrich, PhabricatorCommand],
                  "pipermail": [Pipermail, PipermailOcean, PipermailEnrich, PipermailCommand],
                  "puppetforge": [PuppetForge, PuppetForgeOcean, PuppetForgeEnrich, PuppetForgeCommand],
                  "redmine": [Redmine, RedmineOcean, RedmineEnrich, RedmineCommand],
                  "remo": [ReMo, ReMoOcean, ReMoEnrich, ReMoCommand],
                  "rocketchat": [RocketChat, RocketChatOcean, RocketChatEnrich, RocketChatCommand],
                  "rss": [RSS, RSSOcean, RSSEnrich, RSSCommand],
                  "slack": [Slack, SlackOcean, SlackEnrich, SlackCommand],
                  "stackexchange": [StackExchange, StackExchangeOcean,
                                    StackExchangeEnrich, StackExchangeCommand],
                  "supybot": [Supybot, SupybotOcean, SupybotEnrich, SupybotCommand],
                  "telegram": [Telegram, TelegramOcean, TelegramEnrich, TelegramCommand],
                  "twitter": [Twitter, TwitterOcean, TwitterEnrich, TwitterCommand],
                  "weblate": [Weblate, WeblateOcean, WeblateEnrich, WeblateCommand]
                  }  # Will come from Registry

    for entry_point in pkg_resources.iter_entry_points(ENTRY_POINT_NAME):
        connectors.update(entry_point.load()())
    return connectors


def get_elastic(url, es_index, clean=None, backend=None, es_aliases=None, mapping=None):

    analyzers = None

    if backend:
        backend.set_elastic_url(url)
#        mapping = backend.get_elastic_mappings()
        mapping = backend.mapping
        if hasattr(backend, 'analyzer'):
            analyzers = backend.analyzer
        else:
            analyzers = backend.get_elastic_analyzers()
    try:
        insecure = True
        elastic = ElasticSearch(url=url, index=es_index, mappings=mapping,
                                clean=clean, insecure=insecure,
                                analyzers=analyzers, aliases=es_aliases)

    except ElasticError:
        msg = "Can't connect to Elastic Search. Is it running?"
        logger.error(msg)
        sys.exit(1)

    return elastic


def get_kibiter_version(url):
    """
        Return kibiter major number version

        The url must point to the Elasticsearch used by Kibiter
    """

    config_url = '.kibana/config/_search'
    # Avoid having // in the URL because ES will fail
    if url[-1] != '/':
        url += "/"
    url += config_url
    r = requests.get(url)
    r.raise_for_status()

    if len(r.json()['hits']['hits']) == 0:
        logger.error("Can not get the Kibiter version")
        return None

    version = r.json()['hits']['hits'][0]['_id']
    # 5.4.0-SNAPSHOT
    major_version = version.split(".", 1)[0]
    return major_version


def config_logging(debug):

    if debug:
        logging.basicConfig(level=logging.DEBUG, format='%(asctime)s %(message)s')
        logging.debug("Debug mode activated")
    else:
        logging.basicConfig(level=logging.INFO, format='%(asctime)s %(message)s')
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("requests").setLevel(logging.WARNING)
    # Per commit log is too verbose
    logging.getLogger("perceval.backends.core.git").setLevel(logging.WARNING)


USAGE_MSG = ''
DESC_MSG = ''
EPILOG_MSG = ''


def get_params_parser():
    """Parse command line arguments"""

    parser = argparse.ArgumentParser(usage=USAGE_MSG,
                                     description=DESC_MSG,
                                     epilog=EPILOG_MSG,
                                     formatter_class=argparse.RawDescriptionHelpFormatter,
                                     add_help=False)

    ElasticOcean.add_params(parser)

    parser.add_argument('-h', '--help', action='help',
                        help=argparse.SUPPRESS)
    parser.add_argument('-g', '--debug', dest='debug',
                        action='store_true',
                        help=argparse.SUPPRESS)
    parser.add_argument("--no_incremental", action='store_true',
                        help="don't use last state for data source")
    parser.add_argument("--fetch_cache", action='store_true',
                        help="Use cache for item retrieval")
    parser.add_argument("--enrich", action='store_true',
                        help="Enrich items after retrieving")
    parser.add_argument("--enrich_only", action='store_true',
                        help="Only enrich items (DEPRECATED, use --only-enrich)")
    parser.add_argument("--only-enrich", dest='enrich_only', action='store_true',
                        help="Only enrich items")
    parser.add_argument("--filter-raw", dest='filter_raw',
                        help="Filter raw items. Format: field:value")
    parser.add_argument("--events-enrich", dest='events_enrich', action='store_true',
                        help="Enrich events in items")
    parser.add_argument('--index', help="Ocean index name")
    parser.add_argument('--index-enrich', dest="index_enrich", help="Ocean enriched index name")
    parser.add_argument('--db-user', help="User for db connection (default to root)",
                        default="root")
    parser.add_argument('--db-password', help="Password for db connection (default empty)",
                        default="")
    parser.add_argument('--db-host', help="Host for db connection (default to mariadb)",
                        default="mariadb")
    parser.add_argument('--db-projects-map', help="Projects Mapping DB")
    parser.add_argument('--json-projects-map', help="Projects Mapping JSON file")
    parser.add_argument('--project', help="Project for the repository (origin)")
    parser.add_argument('--refresh-projects', action='store_true', help="Refresh projects in enriched items")
    parser.add_argument('--db-sortinghat', help="SortingHat DB")
    parser.add_argument('--only-identities', action='store_true', help="Only add identities to SortingHat DB")
    parser.add_argument('--refresh-identities', action='store_true', help="Refresh identities in enriched items")
    parser.add_argument('--author_id', nargs='*', help="Field author_ids to be refreshed")
    parser.add_argument('--author_uuid', nargs='*', help="Field author_uuids to be refreshed")
    parser.add_argument('--github-token', help="If provided, github usernames will be retrieved in git enrich.")
    parser.add_argument('--jenkins-rename-file', help="CSV mapping file with nodes renamed schema.")
    parser.add_argument('--studies', action='store_true', help="Execute studies after enrichment.")
    parser.add_argument('--only-studies', action='store_true', help="Execute only studies.")
    parser.add_argument('--bulk-size', default=1000, type=int,
                        help="Number of items per bulk request to Elasticsearch.")
    parser.add_argument('--scroll-wait', default=900, type=int, help="Wait for available scroll (default 900s)")
    parser.add_argument('--scroll-size', default=100, type=int,
                        help="Number of items to get from Elasticsearch when scrolling.")
    parser.add_argument('--pair-programming', action='store_true', help="Do pair programming in git enrich")
    parser.add_argument('--studies-list', nargs='*', help="List of studies to be executed")
    parser.add_argument('backend', help=argparse.SUPPRESS)
    parser.add_argument('backend_args', nargs=argparse.REMAINDER,
                        help=argparse.SUPPRESS)

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)

    return parser


def get_params():
    """ Get params definition from ElasticOcean and from all the backends """

    parser = get_params_parser()
    args = parser.parse_args()

    if not args.enrich_only and not args.only_identities and not args.only_studies:
        if not args.index:
            # Check that the raw index name is defined
            print("[error] --index <name> param is required when collecting items from raw")
            sys.exit(1)

    return args



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Valerio Cosentino <valcos@bitergia.com>
#

import configparser
import datetime
import logging
import random
import string
import sys
import unittest

if '..' not in sys.path:
    sys.path.insert(0, '..')

from base import DB_HOST, DB_SORTINGHAT
from grimoire_elk.utils import get_connectors
from grimoire_elk.enriched.sortinghat_gelk import SortingHat

CONFIG_FILE = 'tests.conf'

logger = logging.getLogger(__name__)


def create_fake_identities():
    identities = []

    for i in range(10):
        username = ''.join(random.choice(string.ascii_lowercase) for _ in range(random.randint(10, 15)))
        email = ''.join(random.choice(string.ascii_lowercase) for _ in range(random.randint(10, 15))) + "@mail.com"
        name = ''.join(random.choice(string.ascii_lowercase) for _ in range(random.randint(10, 20)))

        identity = {'username': username, 'email': email, 'name': name}
        identities.append(identity)

    return identities


class TestLoadIdentities(unittest.TestCase):
    """Tests performance of load identities in GrimoireELK"""

    @classmethod
    def setUpClass(cls):
        config = configparser.ConfigParser()
        config.read(CONFIG_FILE)

        # Sorting hat settings
        cls.db_user = config.get('Database', 'user', fallback='')
        cls.db_password = config.get('Database', 'password', fallback='')
        cls.db_host = config.get('Database', 'host', fallback=DB_HOST)

    def setUp(self):

        # The name of the connector is needed only to get access to the SortingHat DB
        self.enrich_backend = get_connectors()["github"][2](db_sortinghat=DB_SORTINGHAT,
                                                            db_user=self.db_user,
                                                            db_password=self.db_password,
                                                            db_host=self.db_host)

    def _test_load_identities(self, items=10):
        """Test whether fetched items are properly loaded to ES"""

        items_count = 0
        identities_count = 0

        start = datetime.datetime.now().timestamp()
        new_identities = []

        for i in range(items):
            items_count += 1
            identities = create_fake_identities()

            for identity in identities:
                if identity not in new_identities:
                    new_identities.append(identity)

            if items_count % 500 == 0:
                inserted_identities = self.load_bulk_identities(items_count,
                                                                new_identities,
                                                                self.enrich_backend.sh_db,
                                                                self.enrich_backend.get_sh_backend_name())
                identities_count += inserted_identities
                new_identities = []

        if new_identities:
            inserted_identities = self.load_bulk_identities(items_count,
                                                            new_identities,
                                                            self.enrich_backend.sh_db,
                                                            self.enrich_backend.get_sh_backend_name())
            identities_count += inserted_identities

        stop = datetime.datetime.now().timestamp()
        self.assertLess((stop - start), 40.0)

    def load_bulk_identities(self, items_count, new_identities, sh_db, connector_name):
        identities_count = len(new_identities)
        SortingHat.add_identities(sh_db, new_identities, connector_name)

        return identities_count

    def test_load_identities(self):

        self._test_load_identities()


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(message)s')
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("requests").setLevel(logging.WARNING)
    unittest.main(warnings='ignore')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

import logging

from grimoirelab_toolkit.datetime import datetime_utcnow

from .enrich import Enrich, metadata
from ..elastic_mapping import Mapping as BaseMapping

from .utils import get_time_diff_days


logger = logging.getLogger(__name__)


class Mapping(BaseMapping):

    @staticmethod
    def get_elastic_mappings(es_major):
        """Get Elasticsearch mapping.

        :param es_major: major version of Elasticsearch, as string
        :returns:        dictionary with a key, 'items', with the mapping
        """

        mapping = """
        {
            "properties": {
                "description_analyzed": {
                    "type": "text",
                    "index": true
                },
                "subject_analyzed": {
                    "type": "text",
                    "index": true
                }
           }
        } """

        return {"items": mapping}


class RedmineEnrich(Enrich):

    mapping = Mapping

    roles = ['assigned_to', 'author']

    def get_identities(self, item):
        """ Return the identities from an item """

        data = item['data']
        if 'assigned_to' in data:
            user = self.get_sh_identity(data, 'assigned_to')
            yield user
        author = self.get_sh_identity(data, 'author')
        yield author
        # TODO: identities in journals not added yet

    def get_field_author(self):
        return "author"

    def get_sh_identity(self, data, rol):
        identity = {}

        identity['email'] = None
        identity['username'] = None
        identity['name'] = None

        if 'data' in data:
            # data is the full raw item from perceval
            data = data['data']

        if rol + "_data" in data:
            if 'mail' in data[rol + "_data"]:
                identity['email'] = data[rol + "_data"]['mail']
        if rol in data:
            identity['username'] = data[rol]['id']
            identity['name'] = data[rol]['name']

        return identity

    @metadata
    def get_rich_item(self, item):
        eitem = {}

        self.copy_raw_fields(self.RAW_FIELDS_COPY, item, eitem)
        # The real data
        ticket = item['data']

        # data fields to copy
        copy_fields = ["subject", "description", "updated_on",
                       "due_date", "estimated_hours", "done_ratio",
                       "id", "spent_hours", "start_date", "subject",
                       "last_update"]
        for f in copy_fields:
            if f in ticket:
                eitem[f] = ticket[f]
            else:
                eitem[f] = None
        eitem['subject'] = eitem['subject'][:self.KEYWORD_MAX_LENGTH]
        eitem['subject_analyzed'] = eitem['subject']

        eitem['description_analyzed'] = eitem['description']
        if eitem['description']:
            eitem['description'] = eitem['description'][:self.KEYWORD_MAX_LENGTH]

        # Fields which names are translated
        map_fields = {"due_date": "estimated_closing_date",
                      "created_on": "creation_date",
                      "closed_on": "closing_date"}
        for fn in map_fields:
            if fn in ticket:
                eitem[map_fields[fn]] = ticket[fn]
        # Common format
        common = ['category', 'fixed_version', 'priority', 'project', 'status',
                  'tracker', 'author']
        for f in common:
            if f in ticket:
                eitem[f + '_id'] = ticket[f]['id']
                eitem[f + '_name'] = ticket[f]['name']

        len_fields = ['attachments', 'changesets', 'journals', 'relations']
        for f in len_fields:
            if f in ticket:
                eitem[f] = len(ticket[f])

        if 'parent' in ticket:
            eitem['parent_id'] = ticket['parent']['id']

        eitem['url'] = eitem['origin'] + "/issues/" + str(eitem['id'])

        # Time to
        if "closing_date" in eitem:
            eitem['timeopen_days'] = \
                get_time_diff_days(eitem['creation_date'], eitem['closing_date'])
            eitem['timeworking_days'] = \
                get_time_diff_days(eitem['start_date'], eitem['closing_date'])
        else:
            eitem['timeopen_days'] = \
                get_time_diff_days(eitem['creation_date'], datetime_utcnow().replace(tzinfo=None))
            eitem['timeworking_days'] = \
                get_time_diff_days(eitem['start_date'], datetime_utcnow().replace(tzinfo=None))

        eitem.update(self.get_grimoire_fields(item["metadata__updated_on"], "job"))

        if self.sortinghat:
            eitem.update(self.get_item_sh(item, self.roles))

        if self.prjs_map:
            eitem.update(self.get_item_project(eitem))

        self.add_repository_labels(eitem)
        self.add_metadata_filter_raw(eitem)
        return eitem



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

import logging

from grimoirelab_toolkit.datetime import (datetime_utcnow,
                                          str_to_datetime)

from .enrich import Enrich, metadata, SH_UNKNOWN_VALUE
from ..elastic_mapping import Mapping as BaseMapping

from .utils import get_time_diff_days


MAX_SIZE_BULK_ENRICHED_ITEMS = 200
ISSUE_TYPE = 'issue'
COMMENT_TYPE = 'comment'
CLOSED_STATUS_CATEGORY_KEY = 'done'

logger = logging.getLogger(__name__)


class Mapping(BaseMapping):

    @staticmethod
    def get_elastic_mappings(es_major):
        """Get Elasticsearch mapping.

        geopoints type is not created in dynamic mapping

        :param es_major: major version of Elasticsearch, as string
        :returns: dictionary with a key, 'items', with the mapping
        """

        mapping = """
        {
            "properties": {
               "main_description_analyzed": {
                 "type": "text",
                 "index": true
               },
               "releases": {
                 "type": "keyword"
               },
               "body": {
                 "type": "text",
                 "index": true
               }
            }
        }
        """

        return {"items": mapping}


class JiraEnrich(Enrich):

    mapping = Mapping

    roles = ["assignee", "reporter", "creator", "author", "updateAuthor"]

    def get_field_author(self):
        return "reporter"

    def get_sh_identity(self, item, identity_field=None):
        """ Return a Sorting Hat identity using jira user data """

        user = item
        if isinstance(item, dict) and 'data' in item:
            user = item['data']['fields'].get(identity_field, None)
        elif identity_field:
            user = item[identity_field] if 'author' in item else None
        if not user:
            return {}

        identity = {
            'name': None,
            'username': None,
            'email': None
        }

        if 'displayName' in user:
            identity['name'] = user['displayName']
        if 'name' in user:
            identity['username'] = user['name']
        if 'emailAddress' in user:
            identity['email'] = user['emailAddress']

        return identity

    def get_item_sh(self, item, roles=None, date_field=None):
        """Add sorting hat enrichment fields"""

        eitem_sh = {}

        if not self.sortinghat:
            return eitem_sh

        created = str_to_datetime(date_field)

        for rol in roles:
            identity = self.get_sh_identity(item, rol)

            eitem_sh.update(self.get_item_sh_fields(identity, created, rol=rol))

            if not eitem_sh[rol + '_org_name']:
                eitem_sh[rol + '_org_name'] = SH_UNKNOWN_VALUE

            if not eitem_sh[rol + '_name']:
                eitem_sh[rol + '_name'] = SH_UNKNOWN_VALUE

            if not eitem_sh[rol + '_user_name']:
                eitem_sh[rol + '_user_name'] = SH_UNKNOWN_VALUE

        return eitem_sh

    def get_project_repository(self, eitem):
        repo = eitem['origin']
        if eitem['origin'][-1] != "/":
            repo += "/"
        repo += "projects/" + eitem['project_key']
        return repo

    def get_users_data(self, item):
        """ If user fields are inside the global item dict """
        if 'data' in item:
            users_data = item['data']['fields']
        else:
            users_data = item
        return users_data

    def get_identities(self, item):
        """Return the identities from an item"""

        item = item['data']

        if 'fields' not in item:
            return {}

        for field in ["assignee", "reporter", "creator"]:
            if field not in item["fields"]:
                continue
            if item["fields"][field]:
                user = self.get_sh_identity(item["fields"][field])
                yield user

        comments = item.get('comments_data', [])
        for comment in comments:
            if 'author' in comment and comment['author']:
                user = self.get_sh_identity(comment['author'])
                yield user
            if 'updateAuthor' in comment and comment['updateAuthor']:
                user = self.get_sh_identity(comment['updateAuthor'])
                yield user

    @staticmethod
    def fix_value_null(value):
        """Fix <null> values in some Jira parameters.

        In some values fields, as returned by the Jira API,
        some fields appear as <null>. This function convert Them
        to None.

        :param value: string found in a value fields
        :return: same as value, or None
        """
        if value == '<null>':
            return None
        else:
            return value

    @classmethod
    def enrich_fields(cls, fields, eitem):
        """Enrich the fields property of an issue.

        Loops through al properties in issue['fields'],
        using those that are relevant to enrich eitem with new properties.
        Those properties are user defined, depending on options
        configured in Jira. For example, if SCRUM is activated,
        we have a field named "Story Points".

        :param fields: fields property of an issue
        :param eitem: enriched item, which will be modified adding more properties
        """

        for field in fields:
            if field.startswith('customfield_'):
                if type(fields[field]) is dict:
                    if 'name' in fields[field]:
                        if fields[field]['name'] == "Story Points":
                            eitem['story_points'] = fields[field]['value']
                        elif fields[field]['name'] == "Sprint":
                            value = fields[field]['value']
                            if value:
                                sprint = value[0].partition(",name=")[2].split(',')[0]
                                sprint_start = value[0].partition(",startDate=")[2].split(',')[0]
                                sprint_end = value[0].partition(",endDate=")[2].split(',')[0]
                                sprint_complete = value[0].partition(",completeDate=")[2].split(',')[0]
                                eitem['sprint'] = sprint
                                eitem['sprint_start'] = cls.fix_value_null(sprint_start)
                                eitem['sprint_end'] = cls.fix_value_null(sprint_end)
                                eitem['sprint_complete'] = cls.fix_value_null(sprint_complete)

    @metadata
    def get_rich_item(self, item, author_type='creator'):

        eitem = {}

        self.copy_raw_fields(self.RAW_FIELDS_COPY, item, eitem)
        # The real data
        issue = item['data']

        # Fields that are the same in item and eitem
        copy_fields = ["assignee", "reporter"]
        for f in copy_fields:
            if f in issue:
                eitem[f] = issue[f]
            else:
                eitem[f] = None

        eitem['changes'] = issue['changelog']['total']

        if "creator" in issue["fields"] and issue["fields"]["creator"]:
            eitem['creator_name'] = issue["fields"]["creator"].get("displayName", None)
            eitem['creator_login'] = issue["fields"]["creator"].get("name", None)
            if "timeZone" in issue["fields"]["creator"]:
                eitem['creator_tz'] = issue["fields"]["creator"]["timeZone"]

        if "assignee" in issue["fields"] and issue["fields"]["assignee"]:
            eitem['assignee'] = issue["fields"]["assignee"].get("displayName", None)
            if "timeZone" in issue["fields"]["assignee"]:
                eitem['assignee_tz'] = issue["fields"]["assignee"]["timeZone"]

        if 'reporter' in issue['fields'] and issue['fields']['reporter']:
            eitem['reporter_name'] = issue['fields']['reporter'].get('displayName', None)
            eitem['reporter_login'] = issue['fields']['reporter'].get('name', None)
            if "timeZone" in issue["fields"]["reporter"]:
                eitem['reporter_tz'] = issue["fields"]["reporter"]["timeZone"]

        eitem['author_type'] = author_type
        eitem['author_name'] = eitem.get(author_type + '_name', None)
        eitem['author_login'] = eitem.get(author_type + '_login', None)
        eitem['author_tz'] = eitem.get(author_type + '_tz', None)

        eitem['creation_date'] = issue["fields"]['created']

        if 'description' in issue["fields"] and issue["fields"]['description']:
            eitem['main_description'] = issue["fields"]['description'][:self.KEYWORD_MAX_LENGTH]
            eitem['main_description_analyzed'] = issue["fields"]['description']

        eitem['issue_type'] = issue["fields"]['issuetype']['name']
        eitem['issue_description'] = issue["fields"]['issuetype']['description']

        if 'labels' in issue['fields']:
            eitem['labels'] = issue['fields']['labels']

        if 'priority' in issue['fields'] and issue['fields']['priority'] \
                and 'name' in issue['fields']['priority']:
            eitem['priority'] = issue['fields']['priority']['name']

        # data.fields.progress.percent not exists in Puppet JIRA
        if 'progress' in issue['fields']:
            eitem['progress_total'] = issue['fields']['progress']['total']
        eitem['project_id'] = issue['fields']['project']['id']
        eitem['project_key'] = issue['fields']['project']['key']
        eitem['project_name'] = issue['fields']['project']['name']

        if "resolution" in issue['fields'] and issue['fields']['resolution']:
            eitem['resolution_id'] = issue['fields']['resolution']['id']
            eitem['resolution_name'] = issue['fields']['resolution']['name']
            eitem['resolution_description'] = issue['fields']['resolution']['description']
            eitem['resolution_self'] = issue['fields']['resolution']['self']
        eitem['resolution_date'] = issue['fields']['resolutiondate']
        eitem['status_description'] = issue['fields']['status']['description']
        eitem['status'] = issue['fields']['status']['name']
        eitem['status_category_key'] = issue['fields']['status']['statusCategory']['key']
        eitem['is_closed'] = 0
        if eitem['status_category_key'] == CLOSED_STATUS_CATEGORY_KEY:
            eitem['is_closed'] = 1
        eitem['summary'] = issue['fields']['summary']
        if 'timeoriginalestimate' in issue['fields']:
            eitem['original_time_estimation'] = issue['fields']['timeoriginalestimate']
            if eitem['original_time_estimation']:
                eitem['original_time_estimation_hours'] = int(eitem['original_time_estimation']) / 3600
        if 'timespent' in issue['fields']:
            eitem['time_spent'] = issue['fields']['timespent']
            if eitem['time_spent']:
                eitem['time_spent_hours'] = int(eitem['time_spent']) / 3600
        if 'timeestimate' in issue['fields']:
            eitem['time_estimation'] = issue['fields']['timeestimate']
            if eitem['time_estimation']:
                eitem['time_estimation_hours'] = int(eitem['time_estimation']) / 3600
        eitem['watchers'] = issue['fields']['watches']['watchCount']
        eitem['key'] = issue['key']

        # Add extra JSON fields used in Kibana (enriched fields)
        eitem['number_of_comments'] = 0
        eitem['time_to_last_update_days'] = None
        eitem['url'] = None

        # Add id info to allow to coexistence of comments and issues in the same index
        eitem['id'] = '{}_issue_{}_user_{}'.format(eitem['uuid'], issue['id'], author_type)

        if 'comments_data' in issue:
            eitem['number_of_comments'] = len(issue['comments_data'])

        eitem['updated'] = None
        if 'updated' in issue['fields']:
            eitem['updated'] = issue['fields']['updated']

        eitem['url'] = item['origin'] + "/browse/" + issue['key']
        eitem['time_to_close_days'] = \
            get_time_diff_days(issue['fields']['created'], issue['fields']['updated'])
        eitem['time_to_last_update_days'] = \
            get_time_diff_days(issue['fields']['created'], datetime_utcnow().replace(tzinfo=None))

        if 'fixVersions' in issue['fields']:
            eitem['releases'] = []
            for version in issue['fields']['fixVersions']:
                eitem['releases'] += [version['name']]

        self.enrich_fields(issue['fields'], eitem)

        if self.sortinghat:
            eitem.update(self.get_item_sh(item, ["assignee", "reporter", "creator"], issue['fields']['created']))

            # Set the author info to the one identified by author_type
            eitem['author_id'] = eitem[author_type + '_id']
            eitem['author_uuid'] = eitem[author_type + '_uuid']
            eitem['author_name'] = eitem[author_type + '_name']
            eitem['author_user_name'] = eitem[author_type + '_user_name']
            eitem['author_domain'] = eitem[author_type + '_domain']
            eitem['author_gender'] = eitem[author_type + '_gender']
            eitem['author_gender_acc'] = eitem[author_type + '_gender_acc']
            eitem['author_org_name'] = eitem[author_type + '_org_name']
            eitem['author_bot'] = eitem[author_type + '_bot']

            eitem['author_multi_org_names'] = eitem.get(author_type + '_multi_org_names', [])

        if self.prjs_map:
            eitem.update(self.get_item_project(eitem))

        eitem.update(self.get_grimoire_fields(issue['fields']['created'], ISSUE_TYPE))
        eitem["type"] = ISSUE_TYPE

        self.add_repository_labels(eitem)
        self.add_metadata_filter_raw(eitem)
        return eitem

    def get_rich_item_comments(self, comments, eitem):
        ecomments = []

        for comment in comments:
            ecomment = {}

            self.copy_raw_fields(self.RAW_FIELDS_COPY, eitem, ecomment)

            # Copy data from the enriched issue
            ecomment['project_id'] = eitem['project_id']
            ecomment['project_key'] = eitem['project_key']
            ecomment['project_name'] = eitem['project_name']
            ecomment['issue_key'] = eitem['key']
            ecomment['issue_url'] = eitem['url']
            ecomment['issue_type'] = eitem['issue_type']
            ecomment['issue_type'] = eitem['issue_description']

            # Add author and updateAuthor info
            ecomment['author'] = None
            ecomment['updateAuthor'] = None

            if "author" in comment and comment["author"]:
                ecomment['author'] = comment['author']['displayName']
                if "timeZone" in comment["author"]:
                    eitem['author_tz'] = comment["author"]["timeZone"]

            if "updateAuthor" in comment and comment["updateAuthor"]:
                ecomment['updateAuthor'] = comment['updateAuthor']['displayName']
                if "timeZone" in comment["updateAuthor"]:
                    eitem['updateAuthor_tz'] = comment["updateAuthor"]["timeZone"]

            # Add comment-specific data
            ecomment['created'] = str_to_datetime(comment['created']).isoformat()
            ecomment['updated'] = str_to_datetime(comment['updated']).isoformat()
            ecomment['body'] = comment['body']
            ecomment['comment_id'] = comment['id']

            # Add id info to allow to coexistence of comments and issues in the same index
            ecomment['id'] = '{}_comment_{}'.format(eitem['id'], comment['id'])
            ecomment['type'] = COMMENT_TYPE

            if self.sortinghat:
                ecomment.update(self.get_item_sh(comment, ['author', 'updateAuthor'], comment['created']))

            if self.prjs_map:
                ecomment.update(self.get_item_project(ecomment))

            ecomment.update(self.get_grimoire_fields(comment['created'], COMMENT_TYPE))

            self.add_repository_labels(ecomment)
            self.add_metadata_filter_raw(ecomment)
            ecomments.append(ecomment)

        return ecomments

    def get_field_unique_id(self):
        return "id"

    def enrich_items(self, ocean_backend):
        items_to_enrich = []
        num_items = 0
        ins_items = 0

        for item in ocean_backend.fetch():
            # This condition should never happen, since the enriched
            # data heavily relies on the `fields` attribute
            if "fields" not in item["data"]:
                logger.warning("[jira] Skipping item with uuid {}, no fields attribute".format(item['uuid']))
                continue

            eitem_creator = self.get_rich_item(item, author_type='creator')
            eitem_assignee = self.get_rich_item(item, author_type='assignee')
            eitem_reporter = self.get_rich_item(item, author_type='reporter')

            items_to_enrich.append(eitem_creator)
            items_to_enrich.append(eitem_assignee)
            items_to_enrich.append(eitem_reporter)

            comments = item['data'].get('comments_data', [])
            if comments:
                rich_item_comments = self.get_rich_item_comments(comments, eitem_creator)
                items_to_enrich.extend(rich_item_comments)

            if len(items_to_enrich) < MAX_SIZE_BULK_ENRICHED_ITEMS:
                continue

            num_items += len(items_to_enrich)
            ins_items += self.elastic.bulk_upload(items_to_enrich, self.get_field_unique_id())
            items_to_enrich = []

        if len(items_to_enrich) > 0:
            num_items += len(items_to_enrich)
            ins_items += self.elastic.bulk_upload(items_to_enrich, self.get_field_unique_id())

        if num_items != ins_items:
            missing = num_items - ins_items
            logger.error("[jira] {}/{} missing items for Jira".format(missing, num_items))
        else:
            logger.info("[jira] {} items inserted for Jira".format(num_items))

        return num_items



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Animesh Kumar <animuz111@gmail.com>
#

import logging
from datetime import datetime
from grimoirelab_toolkit.datetime import (unixtime_to_datetime,
                                          datetime_utcnow)
from .utils import get_time_diff_days
from .enrich import Enrich, metadata
from ..elastic_mapping import Mapping as BaseMapping

logger = logging.getLogger(__name__)

MAX_SIZE_BULK_ENRICHED_ITEMS = 200
ISSUE_TYPE = 'issue'
COMMENT_TYPE = 'comment'
ISSUE_COMMENT_TYPE = 'issue_comment'


class Mapping(BaseMapping):

    @staticmethod
    def get_elastic_mappings(es_major):
        """Get Elasticsearch mapping.

        :param es_major: major version of Elasticsearch, as string
        :returns:        dictionary with a key, 'items', with the mapping
        """

        mapping = """
        {
            "properties": {
               "title_analyzed": {
                 "type": "text",
                 "index": true
               },
                "body_analyzed": {
                    "type": "text",
                    "index": true
               },
               "id": {
                    "type": "keyword"
               }
            }
        }
        """

        return {"items": mapping}


class PagureEnrich(Enrich):
    mapping = Mapping

    comment_roles = ['user']
    issue_roles = ['user', 'assignee']

    def __init__(self, db_sortinghat=None, json_projects_map=None,
                 db_user='', db_password='', db_host='', db_path=None,
                 db_port=None, db_ssl=False, db_verify_ssl=True, db_tenant=None):
        super().__init__(db_sortinghat=db_sortinghat, json_projects_map=json_projects_map,
                         db_user=db_user, db_password=db_password, db_host=db_host,
                         db_port=db_port, db_path=db_path, db_ssl=db_ssl, db_verify_ssl=db_verify_ssl,
                         db_tenant=db_tenant)

    def set_elastic(self, elastic):
        self.elastic = elastic

    def get_field_author(self):
        return "user"

    def get_field_date(self):
        """ Field with the date in the JSON enriched items """

        return "grimoire_creation_date"

    def get_field_unique_id(self):
        return "id"

    def get_identities(self, item):
        """ Return the identities from an item """

        item = item['data']
        for identity in self.issue_roles:
            if item[identity]:
                user = self.get_sh_identity(item[identity])
                if user:
                    yield user

        # Comments
        if 'comments' in item:
            for comment in item['comments']:
                if 'user' in comment:
                    user = comment.get('user', None)
                    identity = self.get_sh_identity(user)
                    yield identity

    def get_sh_identity(self, item, identity_field=None):
        identity = {}

        user = item  # by default a specific user dict is expected
        if isinstance(item, dict) and 'data' in item:
            user = item['data'][identity_field]
        elif identity_field:
            user = item[identity_field]

        if not user:
            return identity

        identity['email'] = None
        identity['name'] = user.get('fullname', None)
        identity['username'] = user.get('name', None)
        return identity

    def add_gelk_metadata(self, eitem):
        eitem['metadata__gelk_version'] = self.gelk_version
        eitem['metadata__gelk_backend_name'] = self.__class__.__name__
        eitem['metadata__enriched_on'] = datetime_utcnow().isoformat()

    def get_project_repository(self, eitem):
        repo = eitem['origin']
        return repo

    def get_time_to_first_attention(self, item):
        """Get the first date at which a comment was made to the issue by someone
        other than the user who created the issue
        """
        comment_dates = [unixtime_to_datetime(float(comment['date_created'])).isoformat() for comment
                         in item['comments'] if item['user']['name'] != comment['user']['name']]
        if comment_dates:
            return min(comment_dates)
        return None

    def __get_reactions(self, item):
        reactions = {}

        item_reactions = item.get('reactions', {})
        for reaction in item_reactions:
            reactions['reaction_{}'.format(reaction)] = item_reactions[reaction]

        return reactions

    def get_rich_issue_comments(self, comments, eitem):
        ecomments = []

        for comment in comments:
            ecomment = {}

            self.copy_raw_fields(self.RAW_FIELDS_COPY, eitem, ecomment)

            # Copy data from the enriched issue
            ecomment['issue_id'] = eitem['id']
            ecomment['issue_title'] = eitem['title']
            ecomment['issue_status'] = eitem['status']
            ecomment['issue_created_at'] = eitem['created_at']
            ecomment['issue_updated_at'] = eitem['updated_at']
            ecomment['issue_closed_at'] = eitem['closed_at']
            ecomment['issue_pull_requests'] = eitem['related_prs']
            ecomment['item_type'] = COMMENT_TYPE

            # Copy data from the raw comment
            ecomment['body'] = comment['comment'][:self.KEYWORD_MAX_LENGTH]
            ecomment['body_analyzed'] = comment['comment']

            # extract reactions and add it to enriched item
            ecomment.update(self.__get_reactions(comment))

            if 'edited_on' in comment and comment['edited_on']:
                ecomment['comment_updated_at'] = unixtime_to_datetime(float(comment['edited_on'])).isoformat()
                editor = comment['editor']
                ecomment['comment_updated_by_name'] = editor.get('fullname', None)
                ecomment['comment_updated_by_username'] = editor.get('name', None)

            ecomment['comment_created_at'] = unixtime_to_datetime(float(comment['date_created'])).isoformat()
            ecomment['comment_changed_at'] = \
                ecomment['comment_updated_at'] if 'comment_updated_at' in ecomment else ecomment['comment_created_at']
            ecomment['notification'] = comment['notification']
            ecomment['parent'] = comment['parent']

            author = comment.get('user', None)
            if author:
                ecomment['comment_author_username'] = author.get('name', None)
                ecomment['comment_author_name'] = author.get('fullname', None)

            # Add id info to allow to coexistence of items of different types in the same index
            ecomment['id'] = '{}_issue_comment_{}'.format(eitem['id'], comment['id'])
            ecomment.update(self.get_grimoire_fields(ecomment['comment_changed_at'], ISSUE_COMMENT_TYPE))

            if self.sortinghat:
                comment['changed_at'] = ecomment['comment_changed_at']
                ecomment.update(self.get_item_sh(comment, self.comment_roles, 'changed_at'))

            if self.prjs_map:
                ecomment.update(self.get_item_project(ecomment))

            if 'project' in eitem:
                ecomment['project'] = eitem['project']

            self.add_repository_labels(ecomment)
            self.add_metadata_filter_raw(ecomment)
            self.add_gelk_metadata(ecomment)

            ecomments.append(ecomment)

        return ecomments

    def enrich_issue(self, item, eitem):
        eitems = []

        comments = item['data'].get('comments', [])
        if comments:
            rich_item_comments = self.get_rich_issue_comments(comments, eitem)
            eitems.extend(rich_item_comments)

        return eitems

    def enrich_items(self, ocean_backend):
        items_to_enrich = []
        num_items = 0
        ins_items = 0

        for item in ocean_backend.fetch():
            eitem = self.get_rich_item(item)

            if item['category'] == ISSUE_TYPE:
                items_to_enrich.append(eitem)
                eitems = self.enrich_issue(item, eitem)
                items_to_enrich.extend(eitems)

            if len(items_to_enrich) < MAX_SIZE_BULK_ENRICHED_ITEMS:
                continue

            num_items += len(items_to_enrich)
            ins_items += self.elastic.bulk_upload(items_to_enrich, self.get_field_unique_id())
            items_to_enrich = []

        if len(items_to_enrich) > 0:
            num_items += len(items_to_enrich)
            ins_items += self.elastic.bulk_upload(items_to_enrich, self.get_field_unique_id())

        if num_items != ins_items:
            missing = num_items - ins_items
            logger.error("%s/%s missing items for Pagure", str(missing), str(num_items))

        logger.info("%s/%s items inserted for Pagure", str(ins_items), str(num_items))

        return num_items

    @metadata
    def get_rich_item(self, item):

        rich_item = {}
        if item['category'] == 'issue':
            rich_item = self.__get_rich_issue(item)
            self.add_repository_labels(rich_item)
            self.add_metadata_filter_raw(rich_item)
        else:
            logger.error("[pagure] rich item not defined for Pagure category {}".format(item['category']))

        return rich_item

    def __get_rich_issue(self, item):

        rich_issue = {}

        self.copy_raw_fields(self.RAW_FIELDS_COPY, item, rich_issue)

        # The real data
        issue = item['data']

        rich_issue['author_username'] = None
        rich_issue['author_name'] = None
        if 'user' in issue and issue['user']:
            author = issue['user']
            rich_issue['author_username'] = author.get('name', None)
            rich_issue['author_name'] = author.get('fullname', None)

        rich_issue['assignee_username'] = None
        rich_issue['assignee_name'] = None
        if 'assignee' in issue and issue['assignee']:
            assignee = issue['assignee']
            rich_issue['assignee_username'] = assignee.get('name', None)
            rich_issue['assignee_name'] = assignee.get('fullname', None)

        if 'closed_by' in issue and issue['closed_by']:
            closed_by = issue['closed_by']
            rich_issue['closed_by_username'] = closed_by.get('name', None)
            rich_issue['closed_by_name'] = closed_by.get('fullname', None)

        if 'closed_at' in issue and issue['closed_at']:
            rich_issue['closed_at'] = unixtime_to_datetime(float(issue['closed_at'])).isoformat()
        else:
            rich_issue['closed_at'] = None

        rich_issue['id'] = issue['id']
        rich_issue['title'] = issue['title']
        rich_issue['title_analyzed'] = issue['title']
        rich_issue['content'] = issue['content']
        rich_issue['status'] = issue['status']
        rich_issue['close_status'] = issue['close_status']
        rich_issue['num_comments'] = len(issue['comments'])
        rich_issue['created_at'] = unixtime_to_datetime(float(issue['date_created'])).isoformat()
        rich_issue['updated_at'] = unixtime_to_datetime(float(issue['last_updated'])).isoformat()
        rich_issue['blocks'] = issue['blocks']
        rich_issue['private'] = issue['private']
        rich_issue['priority'] = issue['priority']
        rich_issue['milestone'] = issue['milestone']
        rich_issue['related_prs'] = issue['related_prs']
        rich_issue['tags'] = issue['tags']
        rich_issue['custom_fields'] = issue['custom_fields']
        rich_issue['item_type'] = ISSUE_TYPE

        rich_issue['time_to_close_days'] = \
            get_time_diff_days(rich_issue['created_at'], rich_issue['closed_at'])

        if issue['status'] != 'Closed':
            rich_issue['time_open_days'] = \
                get_time_diff_days(rich_issue['created_at'], datetime.utcnow())
        else:
            rich_issue['time_open_days'] = rich_issue['time_to_close_days']

        if self.prjs_map:
            rich_issue.update(self.get_item_project(rich_issue))

        rich_issue['time_to_first_attention'] = None
        if len(issue['comments']) != 0:
            rich_issue['time_to_first_attention'] = \
                get_time_diff_days(rich_issue['created_at'], self.get_time_to_first_attention(issue))

        rich_issue.update(self.get_grimoire_fields(rich_issue['updated_at'], "issue"))

        if self.sortinghat:
            item[self.get_field_date()] = rich_issue[self.get_field_date()]
            rich_issue.update(self.get_item_sh(item, self.issue_roles))

        return rich_issue



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

import logging

from .enrich import Enrich, metadata
from ..elastic_mapping import Mapping as BaseMapping

from grimoirelab_toolkit.datetime import unixtime_to_datetime

logger = logging.getLogger(__name__)


class Mapping(BaseMapping):

    @staticmethod
    def get_elastic_mappings(es_major):
        """Get Elasticsearch mapping.

        :param es_major: major version of Elasticsearch, as string
        :returns:        dictionary with a key, 'items', with the mapping
        """

        mapping = """
        {
            "properties": {
                "message_analyzed": {
                  "type": "text",
                  "fielddata": true,
                  "index": true
                }
           }
        } """

        return {"items": mapping}


class MattermostEnrich(Enrich):
    # This enricher must compatible with the Slack enricher to reuse the Kibiter panel

    mapping = Mapping

    def __init__(self, db_sortinghat=None, json_projects_map=None,
                 db_user='', db_password='', db_host='', db_path=None,
                 db_port=None, db_ssl=False, db_verify_ssl=True, db_tenant=None):
        super().__init__(db_sortinghat=db_sortinghat, json_projects_map=json_projects_map,
                         db_user=db_user, db_password=db_password, db_host=db_host,
                         db_port=db_port, db_path=db_path, db_ssl=db_ssl, db_verify_ssl=db_verify_ssl,
                         db_tenant=db_tenant)

        self.studies = []
        self.studies.append(self.enrich_demography)

    def get_field_author(self):
        return "user_data"

    def get_sh_identity(self, item, identity_field=None):
        identity = {
            'username': None,
            'name': None,
            'email': None
        }

        from_ = item
        if isinstance(item, dict) and 'data' in item:
            if self.get_field_author() not in item['data']:
                # Message from bot
                identity['username'] = item['data']['bot_id']
                return identity
            from_ = item['data'][self.get_field_author()]

        identity['username'] = from_['username']
        identity['name'] = from_['username']

        if 'first_name' in from_:
            name_parts = []
            first_name = from_.get('first_name')
            if first_name:
                name_parts.append(first_name)

            last_name = from_.get('last_name')
            if last_name:
                name_parts.append(last_name)

            composed_name = ' '.join(name_parts)

            identity['name'] = composed_name if composed_name else None
        if 'email' in from_:
            email = from_['email']
            identity['email'] = email if email else None
        return identity

    def get_identities(self, item):
        """ Return the identities from an item """

        identity = self.get_sh_identity(item)
        yield identity

    def get_project_repository(self, eitem):
        # https://chat.openshift.io/8j366ft5affy3p36987pcugaoa
        tokens = eitem['origin'].rsplit("/", 1)
        return tokens[0] + " " + tokens[1]

    @metadata
    def get_rich_item(self, item):
        eitem = {}

        self.copy_raw_fields(self.RAW_FIELDS_COPY, item, eitem)

        # The real data
        message = item['data']

        eitem["reply_count"] = 0  # be sure it is always included
        eitem['message_id'] = message['id']

        # data fields to copy
        copy_fields = ["message", "type", "reply_count", "hashtags", "is_pinned"]
        for f in copy_fields:
            if f in message:
                eitem[f] = message[f]
            else:
                eitem[f] = None

        eitem['message_analyzed'] = eitem['message']

        eitem['reaction_count'] = 0
        eitem['file_count'] = 0
        if 'metadata' in message:
            if 'reactions' in message['metadata']:
                eitem['reaction_count'] = len(message['metadata']['reactions'])
                eitem['reactions'] = self.__get_reactions(message['metadata']['reactions'])
            if 'files' in message['metadata']:
                eitem['file_count'] = len(message['metadata']['files'])
                eitem['files'] = self.__get_files(message['metadata']['files'])

        if 'user_data' in message:
            user_data = message['user_data']
            eitem['roles'] = user_data['roles']
            eitem['position'] = user_data['position']
            eitem['team_id'] = None  # not exists in Mattermost
            timezone = user_data.get('timezone', None)
            if timezone:
                if timezone['useAutomaticTimezone']:
                    eitem['tz'] = timezone['automaticTimezone']
                else:
                    eitem['tz'] = timezone['manualTimezone']
        if 'channel_data' in message:
            channel_data = message['channel_data']
            eitem['channel_name'] = channel_data['display_name']
            eitem['channel_id'] = channel_data['id']
            eitem['channel_create_at'] = unixtime_to_datetime(channel_data['create_at'] / 1000).isoformat()
            eitem['channel_delete_at'] = None if channel_data['delete_at'] == 0 else \
                unixtime_to_datetime(channel_data['delete_at'] / 1000).isoformat()
            eitem['channel_update_at'] = unixtime_to_datetime(channel_data['update_at'] / 1000).isoformat()
            eitem['channel_member_count'] = None
            eitem['channel_message_count'] = channel_data['total_msg_count']
            eitem['channel_team_id'] = channel_data['team_id']

        eitem['is_reply'] = False
        eitem['parent_id'] = message.get('parent_id', message.get('root_id', None))
        if eitem['parent_id']:
            eitem['is_reply'] = True

        eitem = self.__convert_booleans(eitem)

        if self.sortinghat:
            eitem.update(self.get_item_sh(item))

        if self.prjs_map:
            eitem.update(self.get_item_project(eitem))

        eitem.update(self.get_grimoire_fields(item["metadata__updated_on"], "message"))

        self.add_repository_labels(eitem)
        self.add_metadata_filter_raw(eitem)
        return eitem

    def enrich_demography(self, ocean_backend, enrich_backend, alias, date_field="grimoire_creation_date",
                          author_field="author_uuid"):
        super().enrich_demography(ocean_backend, enrich_backend, alias, date_field, author_field=author_field)

    @staticmethod
    def __get_files(message):
        files = []
        for file in message:
            new_file = {
                'file_user_id': file['user_id'],
                'file_post_id': file['post_id'],
                'file_create_at': unixtime_to_datetime(file['create_at'] / 1000).isoformat(),
                'file_update_at': unixtime_to_datetime(file['update_at'] / 1000).isoformat(),
                'file_delete_at': None if file['delete_at'] == 0 else unixtime_to_datetime(
                    file['delete_at'] / 1000).isoformat(),
                'file_name': file['name'],
                'file_extension': file['extension'],
                'file_size': file['size'],
                'file_type': file['mime_type'],
                'file_mini_preview': file['mini_preview']
            }
            files.append(new_file)
        return files

    @staticmethod
    def __get_reactions(message):
        reactions = []
        for reaction in message:
            new_reaction = {
                'reaction_user_id': reaction['user_id'],
                'reaction_post_id': reaction['post_id'],
                'reaction_emoji_name': reaction['emoji_name'],
                'reaction_create_at': unixtime_to_datetime(reaction['create_at'] / 1000).isoformat(),
                'reaction_update_at': unixtime_to_datetime(reaction['update_at'] / 1000).isoformat(),
                'reaction_delete_at': None if reaction['delete_at'] == 0 else unixtime_to_datetime(
                    reaction['delete_at'] / 1000).isoformat()
            }
            reactions.append(new_reaction)
        return reactions

    def __convert_booleans(self, eitem):
        """ Convert True/False to 1/0 for better kibana processing """

        for field in eitem.keys():
            if isinstance(eitem[field], bool):
                if eitem[field]:
                    eitem[field] = 1
                else:
                    eitem[field] = 0
        return eitem  # not needed becasue we are modifying directly the dict



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#   Quan Zhou <quan@bitergia.com>
#

"""Ocean feeder for Elastic from  Perseval data"""


import inspect
import logging

from grimoirelab_toolkit.datetime import unixtime_to_datetime

from datetime import datetime
from ..enriched.utils import get_repository_filter, anonymize_url
from ..elastic_items import ElasticItems
from ..elastic_mapping import Mapping
from ..errors import ELKError
from ..identities.identities import Identities

logger = logging.getLogger(__name__)


PRJ_JSON_FILTER_SEPARATOR = "--filter-"
PRJ_JSON_FILTER_OP_ASSIGNMENT = "="


class ElasticOcean(ElasticItems):

    mapping = Mapping
    identities = Identities

    @classmethod
    def add_params(cls, cmdline_parser):
        """ Shared params in all backends """

        parser = cmdline_parser

        parser.add_argument("-e", "--elastic_url", default="http://127.0.0.1:9200",
                            help="Host with elastic search (default: http://127.0.0.1:9200)")
        parser.add_argument("--elastic_url-enrich",
                            help="Host with elastic search and enriched indexes")

    def __init__(self, perceval_backend, from_date=None, fetch_archive=False,
                 project=None, insecure=True, offset=None, anonymize=False,
                 to_date=None):

        super().__init__(perceval_backend, from_date, insecure, offset, to_date)

        self.fetch_archive = fetch_archive  # fetch from archive
        self.project = project  # project to be used for this data source
        self.anonymize = anonymize

    def set_elastic_url(self, url):
        """ Elastic URL """
        self.elastic_url = url

    def set_elastic(self, elastic):
        """ Elastic used to store last data source state """
        self.elastic = elastic

    def get_field_date(self):
        """ Field with the update in the JSON items. Now the same in all. """
        return "metadata__updated_on"

    def get_field_unique_id(self):
        field = "uuid"
        return field

    def get_elastic_analyzers(self):
        """ Custom analyzers for our indexes  """

        return None

    def get_last_update_from_es(self, filters_=None):
        last_update = self.elastic.get_last_date(self.get_field_date(),
                                                 filters_=filters_)

        return last_update

    def get_connector_name(self):
        """ Find the name for the current connector """
        from ..utils import get_connector_name
        return get_connector_name(type(self))

    @classmethod
    def get_p2o_params_from_url(cls, url):
        """ Get the p2o params given a URL for the data source """

        # if the url doesn't contain a filter separator, return it
        if PRJ_JSON_FILTER_SEPARATOR not in url:
            return {"url": url}

        # otherwise, add the url to the params
        params = {'url': url.split(PRJ_JSON_FILTER_SEPARATOR, 1)[0].strip()}
        # tokenize the filter and add them to the param dict
        tokens = url.split(PRJ_JSON_FILTER_SEPARATOR)[1:]

        if len(tokens) > 1:
            cause = "Too many filters defined for {}, only the first one is considered".format(url)
            logger.warning(cause)

        token = tokens[0]
        filter_tokens = token.split(PRJ_JSON_FILTER_OP_ASSIGNMENT)

        if len(filter_tokens) != 2:
            cause = "Too many tokens after splitting for {} in {}".format(token, url)
            logger.error(cause)
            raise ELKError(cause=cause)

        fltr_name = filter_tokens[0].strip()
        fltr_value = filter_tokens[1].strip()

        params['filter-' + fltr_name] = fltr_value

        return params

    @classmethod
    def get_perceval_params_from_url(cls, url):
        """ Get the perceval params given a URL for the data source """
        return [url]

    def drop_item(self, item):
        """ Drop items not to be inserted in Elastic """
        return False

    def _fix_item(self, item):
        """ Some buggy data sources need fixing (like mbox and message-id) """
        pass

    def add_update_date(self, item):
        """ All item['updated_on'] from perceval is epoch """
        updated = unixtime_to_datetime(item['updated_on'])
        timestamp = unixtime_to_datetime(item['timestamp'])
        item['metadata__updated_on'] = updated.isoformat()
        # Also add timestamp used in incremental enrichment
        item['metadata__timestamp'] = timestamp.isoformat()

    def feed(self, from_date=None, from_offset=None, category=None, branches=None,
             latest_items=None, filter_classified=None, no_update=None, to_date=None):
        """Feed data in Elastic from Perceval"""

        if self.fetch_archive:
            items = self.perceval_backend.fetch_from_archive()
            self.feed_items(items)
            return

        if from_date and from_offset:
            raise RuntimeError("Can't not feed using from_date and from_offset.")

        # We need to filter by repository to support several repositories
        # in the same raw index

        filters_ = [get_repository_filter(self.perceval_backend, self.get_connector_name())]

        # Check if backend supports from_date
        signature = inspect.signature(self.perceval_backend.fetch)

        last_update = None
        if 'from_date' in signature.parameters:
            if from_date:
                last_update = from_date
            else:
                self.last_update = self.get_last_update_from_es(filters_=filters_)
                last_update = self.last_update

            logger.info("[{}] Incremental from: {} until {} for {}".format(
                        self.perceval_backend.__class__.__name__.lower(),
                        last_update, to_date, anonymize_url(self.perceval_backend.origin)))

        offset = None
        if 'offset' in signature.parameters:
            if from_offset:
                offset = from_offset
            else:
                offset = self.elastic.get_last_offset("offset", filters_=filters_)

            if offset is not None:
                logger.info("[{}] Incremental from: {} offset, for {}".format(
                            self.perceval_backend.__class__.__name__.lower(),
                            offset, anonymize_url(self.perceval_backend.origin)))
            else:
                logger.info("[{}] Not incremental".format(
                            self.perceval_backend.__class__.__name__.lower()))

        params = {}
        # category, filter_classified, and to_date params are shared
        # by all Perceval backends
        if category is not None:
            params['category'] = category
        if branches is not None:
            params['branches'] = branches
        if filter_classified is not None:
            params['filter_classified'] = filter_classified
        if to_date:
            params['to_date'] = to_date

        # no_update, latest_items, from_date and offset cannot be used together,
        # thus, the params dictionary is filled with the param available
        # and Perceval is executed
        if no_update:
            params['no_update'] = no_update
            items = self.perceval_backend.fetch(**params)
        elif latest_items:
            params['latest_items'] = latest_items
            items = self.perceval_backend.fetch(**params)
        elif last_update:
            last_update = last_update.replace(tzinfo=None)
            params['from_date'] = last_update
            items = self.perceval_backend.fetch(**params)
        elif offset is not None:
            params['offset'] = offset
            items = self.perceval_backend.fetch(**params)
        else:
            items = self.perceval_backend.fetch(**params)

        self.feed_items(items)
        self.update_items()

    def update_items(self):
        """Perform update operations over a raw index, just after the collection.
        It must be redefined in the raw connectors."""

        return

    def feed_items(self, items):
        task_init = datetime.now()

        items_pack = []  # to feed item in packs
        drop = 0
        added = 0

        for item in items:
            # print("%s %s" % (item['url'], item['lastUpdated_date']))
            # Add date field for incremental analysis if needed
            self.add_update_date(item)
            self._fix_item(item)
            if self.project:
                item['project'] = self.project
            if self.anonymize:
                self.identities.anonymize_item(item)
            if len(items_pack) >= self.elastic.max_items_bulk:
                self._items_to_es(items_pack)
                items_pack = []
            if not self.drop_item(item):
                items_pack.append(item)
                added += 1
            else:
                drop += 1
        self._items_to_es(items_pack)

        total_time_min = (datetime.now() - task_init).total_seconds() / 60

        logger.debug("[{}] Added {} items to index {}".format(
                     self.perceval_backend.__class__.__name__.lower(),
                     added, self.elastic.index))
        logger.debug("[{}] Dropped {} items using drop_item filter".format(
                     self.perceval_backend.__class__.__name__.lower(), drop))
        logger.debug("[{}] Finished in {:.2f} min".format(
                     self.perceval_backend.__class__.__name__.lower(),
                     total_time_min))
        return self

    def _items_to_es(self, json_items):
        """ Append items JSON to ES (data source state) """

        if len(json_items) == 0:
            return

        logger.debug("[{}] Adding items to Raw for {} ({} items)".format(
                     self.perceval_backend.__class__.__name__.lower(),
                     self, len(json_items)))

        field_id = self.get_field_unique_id()

        inserted = self.elastic.bulk_upload(json_items, field_id)

        if len(json_items) != inserted:
            missing = len(json_items) - inserted
            info = json_items[0]

            name = info['backend_name']
            version = info['backend_version']
            origin = info['origin']

            logger.warning("[{}] {}/{} missing JSON items for backend {} [ver. {}], origin {}".format(
                           self.perceval_backend.__class__.__name__.lower(),
                           missing, len(json_items), name, version, origin))

        return inserted



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Alvaro del Castillo <acs@bitergia.com>
#     Valerio Cosentino <valcos@bitergia.com>
#     Quan Zhou <quan@bitergia.com>
#     Miguel Ángel Fernández <mafesan@bitergia.com>
#


import logging
import time
import unittest

from base import TestBaseBackend, DB_SORTINGHAT
from grimoire_elk.enriched.enrich import (logger,
                                          anonymize_url)
from grimoire_elk.enriched.utils import REPO_LABELS


HEADER_JSON = {"Content-Type": "application/json"}


class TestGerrit(TestBaseBackend):
    """Test Gerrit backend"""

    connector = "gerrit"
    ocean_index = "test_" + connector
    enrich_index = "test_" + connector + "_enrich"

    def test_has_identites(self):
        """Test value of has_identities method"""

        enrich_backend = self.connectors[self.connector][2]()
        self.assertTrue(enrich_backend.has_identities())

    def test_items_to_raw(self):
        """Test whether JSON items are properly inserted into ES"""

        result = self._test_items_to_raw()
        self.assertEqual(result['items'], 6)
        self.assertEqual(result['raw'], 6)

    def test_raw_to_enrich(self):
        """Test whether the raw index is properly enriched"""

        result = self._test_raw_to_enrich()
        self.assertEqual(result['raw'], 6)
        self.assertEqual(result['enrich'], 240)

        enrich_backend = self.connectors[self.connector][2]()

        # 1st review
        item = self.items[0]
        eitem = enrich_backend.get_rich_item(item)
        self.assertIn('metadata__enriched_on', eitem)
        self.assertIn('metadata__gelk_backend_name', eitem)
        self.assertIn('metadata__gelk_version', eitem)
        self.assertIn('changeset_status', eitem)
        self.assertIn('changeset_status_value', eitem)
        self.assertIn('changeset_status_value_verified', eitem)
        self.assertIn(REPO_LABELS, eitem)

        self.assertEqual(eitem['time_to_first_review'], 0.06)
        self.assertEqual(eitem['status_value'], '1')
        self.assertEqual(eitem['changeset_status_value_verified'], '1')
        self.assertTrue(eitem['open'])
        self.assertTrue(eitem['wip'])

        comments = item['data']['comments']
        ecomments = enrich_backend.get_rich_item_comments(comments, eitem)
        self.assertEqual(len(ecomments), 46)

        for ecomment in ecomments:
            self.assertIn('comment_message', ecomment)
            self.assertIn('comment_message_analyzed', ecomment)
            self.assertIn('metadata__enriched_on', ecomment)
            self.assertIn('metadata__gelk_backend_name', ecomment)
            self.assertIn('metadata__gelk_version', ecomment)
            self.assertIn(REPO_LABELS, ecomment)
            self.assertTrue(ecomment['open'])
            self.assertTrue(ecomment['wip'])

        patchsets = item['data']['patchSets']
        epatchsets = enrich_backend.get_rich_item_patchsets(patchsets, eitem)
        self.assertEqual(len(epatchsets), 19)

        for epatchset in epatchsets:
            self.assertIn('metadata__enriched_on', epatchset)
            self.assertIn('metadata__gelk_backend_name', epatchset)
            self.assertIn('metadata__gelk_version', epatchset)
            self.assertIn('changeset_status', epatchset)
            self.assertIn('changeset_status_value', epatchset)
            self.assertIn('changeset_status_value_verified', epatchset)
            self.assertIn(REPO_LABELS, epatchset)
            self.assertTrue(epatchset['open'])
            self.assertTrue(epatchset['wip'])

        eapprovals = [ei for ei in epatchsets if 'is_gerrit_approval' in ei]
        epatchsets = [ei for ei in epatchsets if 'is_gerrit_patchset' in ei]

        self.assertEqual(len(epatchsets), 5)

        expected_patchset_time_to_first_review = [0.06, 5.21, None, 0.37, None]
        for i, epatchset in enumerate(epatchsets):
            self.assertIn('metadata__enriched_on', epatchset)
            self.assertIn('metadata__gelk_backend_name', epatchset)
            self.assertIn('metadata__gelk_version', epatchset)
            self.assertIn('changeset_status', epatchset)
            self.assertIn('changeset_status_value', epatchset)
            self.assertIn('changeset_status_value_verified', epatchset)
            self.assertIn(REPO_LABELS, epatchset)
            self.assertEqual(epatchset['patchset_time_to_first_review'], expected_patchset_time_to_first_review[i])
            self.assertTrue(epatchset['open'])
            self.assertTrue(epatchset['wip'])

        self.assertEqual(len(eapprovals), 14)
        for eapproval in eapprovals:
            self.assertIn('approval_description', eapproval)
            self.assertIn('approval_description_analyzed', eapproval)
            self.assertIn('changeset_status', eapproval)
            self.assertIn('changeset_status_value', eapproval)
            self.assertIn('changeset_status_value_verified', eapproval)
            self.assertIn('metadata__enriched_on', eapproval)
            self.assertIn('metadata__gelk_backend_name', eapproval)
            self.assertIn('metadata__gelk_version', eapproval)
            self.assertIn(REPO_LABELS, eapproval)
            self.assertTrue(eapproval['open'])
            self.assertTrue(eapproval['wip'])

        # 2nd review
        item = self.items[1]
        eitem = enrich_backend.get_rich_item(item)
        self.assertIn('metadata__enriched_on', eitem)
        self.assertIn('metadata__gelk_backend_name', eitem)
        self.assertIn('metadata__gelk_version', eitem)
        self.assertIn(REPO_LABELS, eitem)
        self.assertTrue(eitem['open'])
        self.assertTrue(eitem['wip'])

        comments = item['data']['comments']
        ecomments = enrich_backend.get_rich_item_comments(comments, eitem)
        self.assertEqual(len(ecomments), 23)

        patchsets = item['data']['patchSets']
        epatchsets = enrich_backend.get_rich_item_patchsets(patchsets, eitem)
        self.assertEqual(len(epatchsets), 11)

        # 3rd review
        item = self.items[2]
        eitem = enrich_backend.get_rich_item(item)
        self.assertIn('metadata__enriched_on', eitem)
        self.assertIn('metadata__gelk_backend_name', eitem)
        self.assertIn('metadata__gelk_version', eitem)
        self.assertIn(REPO_LABELS, eitem)
        self.assertTrue(eitem['open'])
        self.assertFalse(eitem['wip'])

        comments = item['data']['comments']
        ecomments = enrich_backend.get_rich_item_comments(comments, eitem)
        self.assertEqual(len(ecomments), 24)

        patchsets = item['data']['patchSets']
        epatchsets = enrich_backend.get_rich_item_patchsets(patchsets, eitem)
        self.assertEqual(len(epatchsets), 17)

        # 4th review
        item = self.items[3]
        eitem = enrich_backend.get_rich_item(item)
        self.assertIn('metadata__enriched_on', eitem)
        self.assertIn('metadata__gelk_backend_name', eitem)
        self.assertIn('metadata__gelk_version', eitem)
        self.assertIn(REPO_LABELS, eitem)
        self.assertTrue(eitem['open'])
        self.assertFalse(eitem['wip'])

        comments = item['data']['comments']
        ecomments = enrich_backend.get_rich_item_comments(comments, eitem)
        self.assertEqual(len(ecomments), 6)

        patchsets = item['data']['patchSets']
        epatchsets = enrich_backend.get_rich_item_patchsets(patchsets, eitem)
        self.assertEqual(len(epatchsets), 3)

        # 5th review
        item = self.items[4]
        eitem = enrich_backend.get_rich_item(item)
        self.assertIn('metadata__enriched_on', eitem)
        self.assertIn('metadata__gelk_backend_name', eitem)
        self.assertIn('metadata__gelk_version', eitem)
        self.assertIn(REPO_LABELS, eitem)
        self.assertTrue(eitem['open'])
        self.assertFalse(eitem['wip'])

        comments = item['data']['comments']
        ecomments = enrich_backend.get_rich_item_comments(comments, eitem)
        self.assertEqual(len(ecomments), 31)

        patchsets = item['data']['patchSets']
        epatchsets = enrich_backend.get_rich_item_patchsets(patchsets, eitem)
        self.assertEqual(len(epatchsets), 6)

        # 6th review
        item = self.items[5]
        eitem = enrich_backend.get_rich_item(item)
        self.assertIn('metadata__enriched_on', eitem)
        self.assertIn('metadata__gelk_backend_name', eitem)
        self.assertIn('metadata__gelk_version', eitem)
        self.assertIn(REPO_LABELS, eitem)
        self.assertTrue(eitem['open'])
        self.assertFalse(eitem['wip'])

        comments = item['data']['comments']
        ecomments = enrich_backend.get_rich_item_comments(comments, eitem)
        self.assertEqual(len(ecomments), 46)

        patchsets = item['data']['patchSets']
        epatchsets = enrich_backend.get_rich_item_patchsets(patchsets, eitem)
        self.assertEqual(len(epatchsets), 2)

    def test_demography_study(self):
        """ Test that the demography study works correctly """

        alias = "demographics"
        study, ocean_backend, enrich_backend = self._test_study('enrich_demography')

        with self.assertLogs(logger, level='INFO') as cm:

            if study.__name__ == "enrich_demography":
                study(ocean_backend, enrich_backend, alias, date_field="grimoire_creation_date")

            self.assertEqual(cm.output[0],
                             'INFO:grimoire_elk.enriched.enrich:[gerrit] '
                             'Demography starting study %s/test_gerrit_enrich'
                             % anonymize_url(self.es_con))
            self.assertEqual(cm.output[-1],
                             'INFO:grimoire_elk.enriched.enrich:[gerrit] Demography end %s/test_gerrit_enrich'
                             % anonymize_url(self.es_con))

        time.sleep(5)  # HACK: Wait until git enrich index has been written
        items = [i for i in enrich_backend.fetch()]
        for item in items:
            if item['type'] == 'patchset' and item['patchset_author_name'] is None:
                self.assertFalse('demography_min_date' in item.keys())
                self.assertFalse('demography_max_date' in item.keys())
            elif 'reviewer_uuid' in item:
                self.assertTrue('demography_min_date' in item.keys())
                self.assertTrue('demography_max_date' in item.keys())

        r = enrich_backend.elastic.requests.get(enrich_backend.elastic.index_url + "/_alias",
                                                headers=HEADER_JSON, verify=False)
        self.assertIn(alias, r.json()[enrich_backend.elastic.index]['aliases'])

    def test_demography_contribution_study(self):
        """ Test that the demography contribution study works correctly """

        alias = "demographics_contribution"
        study, ocean_backend, enrich_backend = self._test_study('enrich_demography_contribution')

        with self.assertLogs(logger, level='INFO') as cm:

            if study.__name__ == "enrich_demography_contribution":
                study(ocean_backend, enrich_backend, alias, date_field="grimoire_creation_date")

            self.assertEqual(cm.output[0],
                             'INFO:grimoire_elk.enriched.enrich:[gerrit] '
                             'Demography Contribution starting study %s/test_gerrit_enrich'
                             % anonymize_url(self.es_con))
            self.assertEqual(cm.output[-1],
                             'INFO:grimoire_elk.enriched.enrich:[gerrit] Demography Contribution end %s/test_gerrit_enrich'
                             % anonymize_url(self.es_con))

        time.sleep(5)  # HACK: Wait until gerrit enrich index has been written
        items = [i for i in enrich_backend.fetch()]
        for item in items:
            if 'author_uuid' not in item:
                continue
            field_type = item['type']
            if field_type == 'approval':
                self.assertIn('approval_min_date', item)
                self.assertIn('approval_max_date', item)
            elif field_type == 'changeset':
                self.assertIn('changeset_min_date', item)
                self.assertIn('changeset_max_date', item)
            elif field_type == 'comment':
                self.assertIn('comment_min_date', item)
                self.assertIn('comment_max_date', item)
            elif field_type == 'patchset':
                self.assertIn('patchset_min_date', item)
                self.assertIn('patchset_max_date', item)

        r = enrich_backend.elastic.requests.get(enrich_backend.elastic.index_url + "/_alias",
                                                headers=HEADER_JSON, verify=False)
        self.assertIn(alias, r.json()[enrich_backend.elastic.index]['aliases'])

    def test_raw_to_enrich_sorting_hat(self):
        """Test enrich with SortingHat"""

        result = self._test_raw_to_enrich(sortinghat=True)
        self.assertEqual(result['raw'], 6)
        self.assertEqual(result['enrich'], 240)

        enrich_backend = self.connectors[self.connector][2](db_sortinghat=DB_SORTINGHAT,
                                                            db_user=self.db_user,
                                                            db_password=self.db_password)

        for item in self.items:
            eitem = enrich_backend.get_rich_item(item)

            self.assertIn('changeset_author_id', eitem)
            self.assertIn('changeset_author_uuid', eitem)
            self.assertIn('changeset_author_name', eitem)
            self.assertIn('changeset_author_user_name', eitem)
            self.assertIn('changeset_author_domain', eitem)
            self.assertIn('changeset_author_gender', eitem)
            self.assertIn('changeset_author_gender_acc', eitem)
            self.assertIn('changeset_author_org_name', eitem)
            self.assertIn('changeset_author_bot', eitem)
            self.assertIn('changeset_author_multi_org_names', eitem)

            comments = item['data']['comments']
            ecomments = enrich_backend.get_rich_item_comments(comments, eitem)

            for ecomment in ecomments:
                self.assertIn('changeset_author_id', ecomment)
                self.assertIn('changeset_author_uuid', ecomment)
                self.assertIn('changeset_author_name', ecomment)
                self.assertIn('changeset_author_user_name', ecomment)
                self.assertIn('changeset_author_domain', ecomment)
                self.assertIn('changeset_author_gender', ecomment)
                self.assertIn('changeset_author_gender_acc', ecomment)
                self.assertIn('changeset_author_org_name', ecomment)
                self.assertIn('changeset_author_bot', ecomment)
                self.assertIn('changeset_author_multi_org_names', ecomment)

            patchsets = item['data']['patchSets']
            epatchsets = enrich_backend.get_rich_item_patchsets(patchsets, eitem)

            for epatchset in epatchsets:
                self.assertIn('changeset_author_id', epatchset)
                self.assertIn('changeset_author_uuid', epatchset)
                self.assertIn('changeset_author_name', epatchset)
                self.assertIn('changeset_author_user_name', epatchset)
                self.assertIn('changeset_author_domain', epatchset)
                self.assertIn('changeset_author_gender', epatchset)
                self.assertIn('changeset_author_gender_acc', epatchset)
                self.assertIn('changeset_author_org_name', epatchset)
                self.assertIn('changeset_author_bot', epatchset)
                self.assertIn('changeset_author_multi_org_names', epatchset)

    def test_raw_to_enrich_projects(self):
        """Test enrich with Projects"""

        result = self._test_raw_to_enrich(projects=True)
        self.assertEqual(result['raw'], 6)
        self.assertEqual(result['enrich'], 240)

    def test_copy_raw_fields(self):
        """Test copied raw fields"""

        self._test_raw_to_enrich()
        enrich_backend = self.connectors[self.connector][2]()

        for item in self.items:
            eitem = enrich_backend.get_rich_item(item)
            for attribute in enrich_backend.RAW_FIELDS_COPY:
                if attribute in item:
                    self.assertEqual(item[attribute], eitem[attribute])
                else:
                    self.assertIsNone(eitem[attribute])

    def test_refresh_identities(self):
        """Test refresh identities"""

        result = self._test_refresh_identities()
        # ... ?


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(message)s')
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("requests").setLevel(logging.WARNING)
    unittest.main(warnings='ignore')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Valerio Cosentino <valcos@bitergia.com>
#   Nishchith Shetty <inishchith@gmail.com>
#

import logging

import hashlib

from dateutil.relativedelta import relativedelta

from elasticsearch import Elasticsearch as ES, RequestsHttpConnection

from .enrich import (Enrich,
                     metadata)
from .graal_study_evolution import (get_to_date,
                                    get_unique_repository,
                                    get_files_at_time)
from .utils import fix_field_date, anonymize_url
from ..elastic_mapping import Mapping as BaseMapping

from grimoirelab_toolkit.datetime import datetime_utcnow
from grimoire_elk.elastic import ElasticSearch

MAX_SIZE_BULK_ENRICHED_ITEMS = 200
LANGUAGES = {
    'java': 'Java',
    'py': 'Python',
    'php': 'PHP',
    'scala': 'Scala',
    'js': 'JavaScript',
    'rb': 'Ruby',
    'cs': 'C#',
    'cpp': 'C/C++',
    'cc': 'C/C++',
    'c': 'C/C++',
    'h': 'C/C++',
    'hpp': 'C/C++',
    'hs': 'Haskell',
    'lua': 'Lua',
    'go': 'Go',
    'swift': 'Swift',
    'm': 'Objective-C',
    'mm': 'Objective-C',
    'py': 'Python',
    'lisp': 'Lisp',
    'pl': 'Perl',
    'pm': 'Perl',
    'r': 'R',
    'sh': 'Shell',
    'kt': 'Kotlin',
    'kts': 'Kotlin',
    'xml': 'XML',
    'yml': 'YAML',
    'json': 'JSON',
    'md': 'Markdown',
    'gradle': 'Gradle',
    'dockerfile': 'Docker',
    'sql': 'SQL',
    'ts': 'TypeScript'
}

logger = logging.getLogger(__name__)


class Mapping(BaseMapping):

    @staticmethod
    def get_elastic_mappings(es_major):
        """Get Elasticsearch mapping.

        Ensure data.message is string, since it can be very large

        :param es_major: major version of Elasticsearch, as string
        :returns:        dictionary with a key, 'items', with the mapping
        """

        mapping = '''
         {
            "dynamic":true,
            "properties": {
                "id" : {
                    "type" : "keyword"
                },
                "interval_months" : {
                    "type" : "long"
                },
                "origin" : {
                    "type" : "keyword"
                },
                "study_creation_date" : {
                    "type" : "date"
                },
                "total_blanks" : {
                    "type" : "long"
                },
                "total_blanks_per_loc" : {
                    "type" : "float"
                },
                "total_ccn" : {
                    "type" : "long"
                },
                "total_comments" : {
                    "type" : "long"
                },
                "total_comments_per_loc" : {
                    "type" : "float"
                },
                "total_files" : {
                    "type" : "long"
                },
                "total_loc" : {
                    "type" : "long"
                },
                "total_loc_per_function" : {
                    "type" : "float"
                },
                "total_num_funs" : {
                    "type" : "long"
                },
                "total_tokens" : {
                    "type" : "long"
                }
            }
        }
        '''

        return {"items": mapping}


class CocomEnrich(Enrich):
    metrics = ["ccn", "num_funs", "tokens", "loc", "comments", "blanks"]

    def __init__(self, db_sortinghat=None, json_projects_map=None,
                 db_user='', db_password='', db_host='', db_path=None,
                 db_port=None, db_ssl=False, db_verify_ssl=True, db_tenant=None):
        super().__init__(db_sortinghat=db_sortinghat, json_projects_map=json_projects_map,
                         db_user=db_user, db_password=db_password, db_host=db_host,
                         db_port=db_port, db_path=db_path, db_ssl=db_ssl, db_verify_ssl=db_verify_ssl,
                         db_tenant=db_tenant)

        self.studies = []
        self.studies.append(self.enrich_cocom_analysis)

    def get_identities(self, item):
        """ Return the identities from an item """
        identities = []

        return identities

    def has_identities(self):
        """ Return whether the enriched items contains identities """

        return False

    def get_field_unique_id(self):
        return "id"

    def extract_modules(self, file_path):
        """ Extracts module path from the given file path """
        path_chunks = file_path.split('/')

        modules = []
        for idx in range(len(path_chunks)):
            sub_path = '/'.join(path_chunks[:idx])

            if sub_path:
                modules.append(sub_path)

        return modules

    def get_language(self, ext):
        language = None
        if ext:
            language = LANGUAGES.get(ext.lower(), ext.upper())

        return language

    @metadata
    def get_rich_item(self, file_analysis):
        eitem = {}
        for metric in self.metrics:
            eitem[metric] = file_analysis.get(metric, None)
        eitem["file_path"] = file_analysis.get("file_path", None)
        eitem['modules'] = self.extract_modules(eitem['file_path'])
        eitem = self.__add_derived_metrics(file_analysis, eitem)
        # Get the name of the file if not extension
        eitem["ext"] = file_analysis["ext"].split("/")[-1] if "ext" in file_analysis else None
        eitem['language'] = self.get_language(eitem["ext"])

        return eitem

    def get_rich_items(self, item):
        # The real data
        entry = item['data']

        enriched_items = []

        for file_analysis in entry["analysis"]:
            eitem = self.get_rich_item(file_analysis)

            self.copy_raw_fields(self.RAW_FIELDS_COPY, item, eitem)

            # common attributes
            eitem['commit_sha'] = entry['commit']
            eitem['commit_parents'] = entry.get('parents', None)
            eitem['files_at_commit'] = entry.get('files', None)
            eitem['author'] = entry['Author']
            eitem['committer'] = entry['Commit']
            eitem['message'] = entry.get('message', None)
            eitem['author_date'] = fix_field_date(entry['AuthorDate'])
            eitem['commit_date'] = fix_field_date(entry['CommitDate'])

            # Other enrichment
            eitem["repo_url"] = item["origin"]
            if eitem["repo_url"].startswith('http'):
                eitem["repo_url"] = anonymize_url(eitem["repo_url"])

            if self.prjs_map:
                eitem.update(self.get_item_project(eitem))

            # uuid
            eitem['id'] = "{}_{}".format(eitem['commit_sha'], eitem['file_path'])

            eitem.update(self.get_grimoire_fields(entry["AuthorDate"], "file"))

            self.add_repository_labels(eitem)
            self.add_metadata_filter_raw(eitem)

            enriched_items.append(eitem)

        return enriched_items

    def __add_derived_metrics(self, file_analysis, eitem):
        """ Add derived metrics fields """

        # TODO: Fix Logic: None rather than 1
        if eitem["loc"] is not None and eitem["comments"] is not None and eitem["num_funs"] is not None:
            eitem["comments_per_loc"] = round(eitem["comments"] / max(eitem["loc"], 1), 2)
            eitem["blanks_per_loc"] = round(eitem["blanks"] / max(eitem["loc"], 1), 2)
            eitem["loc_per_function"] = round(eitem["loc"] / max(eitem["num_funs"], 1), 2)
        else:
            eitem["comments_per_loc"] = None
            eitem["blanks_per_loc"] = None
            eitem["loc_per_function"] = None

        return eitem

    def enrich_items(self, ocean_backend, events=False):
        items_to_enrich = []
        num_items = 0
        ins_items = 0

        for item in ocean_backend.fetch():
            rich_items = self.get_rich_items(item)

            items_to_enrich.extend(rich_items)
            if len(items_to_enrich) < MAX_SIZE_BULK_ENRICHED_ITEMS:
                continue

            num_items += len(items_to_enrich)
            ins_items += self.elastic.bulk_upload(items_to_enrich, self.get_field_unique_id())
            items_to_enrich = []

        if len(items_to_enrich) > 0:
            num_items += len(items_to_enrich)
            ins_items += self.elastic.bulk_upload(items_to_enrich, self.get_field_unique_id())

        if num_items != ins_items:
            missing = num_items - ins_items
            logger.error("[cocom] {}/{} missing items".format(
                         missing, num_items))
        else:
            logger.info("[cocom] {} items inserted".format(
                        num_items))

        return num_items

    def enrich_cocom_analysis(self, ocean_backend, enrich_backend, no_incremental=False,
                              out_index="cocom_enrich_graal_repo", interval_months=[3],
                              date_field="grimoire_creation_date"):

        logger.info("[cocom] study enrich-cocom-analysis start")

        es_in = ES([enrich_backend.elastic_url], retry_on_timeout=True, timeout=100,
                   verify_certs=self.elastic.requests.verify, connection_class=RequestsHttpConnection)
        in_index = enrich_backend.elastic.index
        interval_months = list(map(int, interval_months))

        unique_repos = es_in.search(
            index=in_index,
            body=get_unique_repository())

        repositories = [repo['key'] for repo in unique_repos['aggregations']['unique_repos'].get('buckets', [])]
        current_month = datetime_utcnow().replace(day=1, hour=0, minute=0, second=0)

        logger.info("[cocom] study enrich-cocom-analysis {} repositories to process".format(
                    len(repositories)))
        es_out = ElasticSearch(enrich_backend.elastic.url, out_index, mappings=Mapping)
        es_out.add_alias("cocom_study")

        num_items = 0
        ins_items = 0

        for repository_url in repositories:
            repository_url_anonymized = repository_url
            if repository_url_anonymized.startswith('http'):
                repository_url_anonymized = anonymize_url(repository_url_anonymized)

            logger.info("[cocom] study enrich-cocom-analysis start analysis for {}".format(
                        repository_url_anonymized))
            evolution_items = []

            for interval in interval_months:

                to_month = get_to_date(es_in, in_index, out_index, repository_url, interval)
                to_month = to_month.replace(month=int(interval), day=1, hour=0, minute=0, second=0)

                while to_month < current_month:
                    files_at_time = es_in.search(
                        index=in_index,
                        body=get_files_at_time(repository_url, to_month.isoformat())
                    )['aggregations']['file_stats'].get("buckets", [])

                    if not len(files_at_time):
                        to_month = to_month + relativedelta(months=+interval)
                        continue

                    repository_name = repository_url.split("/")[-1]
                    total_per_lang = {}

                    for file_ in files_at_time:
                        file_details = file_["1"]["hits"]["hits"][0]["_source"]

                        if "language" in file_details:
                            lang = file_details["language"]
                            total_per_lang[lang] = total_per_lang.get(lang, {})

                            for metric in self.metrics:
                                total_per_lang[lang][metric] = total_per_lang[lang].get(metric, 0)
                                total_per_lang[lang][metric] += file_details[metric] if file_details[metric] is not None else 0

                            total_per_lang[lang]["total_files"] = total_per_lang[lang].get("total_files", 0) + 1

                    for language in total_per_lang:
                        total = total_per_lang[language]
                        if total["loc"] > 0:
                            hash_repo_url = hashlib.md5(repository_url_anonymized.encode('utf-8')).hexdigest()
                            to_month_iso = to_month.isoformat()
                            evolution_item = {
                                "id": "{}_{}_{}_{}".format(to_month_iso, hash_repo_url, interval, language),
                                "repo_url": repository_url_anonymized,
                                "origin": repository_url,
                                "interval_months": interval,
                                "study_creation_date": to_month_iso,
                                "language": language,
                                "total_files": total["total_files"]
                            }

                            for metric in self.metrics:
                                evolution_item["total_" + metric] = total[metric]

                            evolution_item["total_comments_per_loc"] = round(
                                evolution_item["total_comments"] / max(evolution_item["total_loc"], 1), 2)
                            evolution_item["total_blanks_per_loc"] = round(
                                evolution_item["total_blanks"] / max(evolution_item["total_loc"], 1), 2)
                            evolution_item["total_loc_per_function"] = round(
                                evolution_item["total_loc"] / max(evolution_item["total_num_funs"], 1), 2)

                            evolution_item.update(self.get_grimoire_fields(evolution_item["study_creation_date"], "stats"))
                            evolution_items.append(evolution_item)

                    if len(evolution_items) >= self.elastic.max_items_bulk:
                        num_items += len(evolution_items)
                        ins_items += es_out.bulk_upload(evolution_items, self.get_field_unique_id())
                        evolution_items = []

                    to_month = to_month + relativedelta(months=+interval)

                if len(evolution_items) > 0:
                    num_items += len(evolution_items)
                    ins_items += es_out.bulk_upload(evolution_items, self.get_field_unique_id())

                if num_items != ins_items:
                    missing = num_items - ins_items
                    logger.error(
                        "[cocom] study enrich-cocom-analysis {}/{} missing items for Graal CoCom Analysis "
                        "Study".format(missing, num_items)
                    )
                else:
                    logger.info(
                        "[cocom] study enrich-cocom-analysis {} items inserted for Graal CoCom Analysis "
                        "Study".format(num_items)
                    )

            logger.info(
                "[cocom] study enrich-cocom-analysis End analysis for {} with month interval".format(
                    repository_url_anonymized)
            )

        logger.info("[cocom] study enrich-cocom-analysis End")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Nishchith Shetty <inishchith@gmail.com>
#
import logging
import unittest

from base import TestBaseBackend
from grimoire_elk.raw.graal import GraalOcean
from grimoire_elk.enriched.cocom import logger


HEADER_JSON = {"Content-Type": "application/json"}


class TestCoCom(TestBaseBackend):
    """Test CoCom backend"""

    connector = "cocom"
    ocean_index = "test_" + connector
    enrich_index = "test_" + connector + "_enrich"

    def test_has_identites(self):
        """Test value of has_identities method"""

        enrich_backend = self.connectors[self.connector][2]()
        self.assertFalse(enrich_backend.has_identities())

    def test_items_to_raw(self):
        """Test whether JSON items are properly inserted into ES"""

        result = self._test_items_to_raw()

        self.assertGreater(result['items'], 0)
        self.assertGreater(result['raw'], 0)
        self.assertGreaterEqual(result['items'], result['raw'])

    def test_raw_to_enrich(self):
        """Test whether the raw index is properly enriched"""

        result = self._test_raw_to_enrich()

        self.assertGreater(result['raw'], 0)
        self.assertGreater(result['enrich'], 0)
        self.assertGreaterEqual(result['enrich'], result['raw'])

        enrich_backend = self.connectors[self.connector][2]()

        item = self.items[0]
        eitem = enrich_backend.get_rich_items(item)[0]
        self.assertEqual(eitem['ccn'], 75)
        self.assertEqual(eitem['num_funs'], 31)
        self.assertEqual(eitem['tokens'], 2207)
        self.assertEqual(eitem['loc'], 372)
        self.assertEqual(eitem['ext'], "py")
        self.assertEqual(eitem['blanks'], 158)
        self.assertEqual(eitem['comments'], 193)
        self.assertEqual(eitem['file_path'], "graal/graal.py")
        self.assertEqual(eitem['modules'], ["graal"])
        self.assertEqual(eitem["comments_per_loc"], 0.52)
        self.assertEqual(eitem["blanks_per_loc"], 0.42)
        self.assertEqual(eitem["loc_per_function"], 12.0)
        self.assertIn('commit_parents', eitem)
        self.assertIn('files_at_commit', eitem)

        item = self.items[1]
        eitem = enrich_backend.get_rich_items(item)[0]
        self.assertEqual(eitem['ccn'], 70)
        self.assertEqual(eitem['num_funs'], 52)
        self.assertEqual(eitem['tokens'], 4623)
        self.assertEqual(eitem['loc'], 527)
        self.assertEqual(eitem['ext'], "py")
        self.assertEqual(eitem['blanks'], 204)
        self.assertEqual(eitem['comments'], 77)
        self.assertEqual(eitem['file_path'], "tests/test_graal.py")
        self.assertEqual(eitem['modules'], ["tests"])
        self.assertEqual(eitem["comments_per_loc"], 0.15)
        self.assertEqual(eitem["blanks_per_loc"], 0.39)
        self.assertEqual(eitem["loc_per_function"], 10.13)
        self.assertIn('commit_parents', eitem)
        self.assertIn('files_at_commit', eitem)

        item = self.items[2]
        eitem = enrich_backend.get_rich_items(item)[0]
        self.assertEqual(eitem['ccn'], 8)
        self.assertEqual(eitem['num_funs'], 3)
        self.assertEqual(eitem['tokens'], 421)
        self.assertEqual(eitem['loc'], 80)
        self.assertEqual(eitem['ext'], "py")
        self.assertEqual(eitem['blanks'], 26)
        self.assertEqual(eitem['comments'], 63)
        self.assertEqual(eitem['file_path'], "graal/backends/core/analyzers/lizard.py")
        self.assertEqual(eitem['modules'], ["graal", "graal/backends", "graal/backends/core", "graal/backends/core/analyzers"])
        self.assertEqual(eitem["comments_per_loc"], 0.79)
        self.assertEqual(eitem["blanks_per_loc"], 0.33)
        self.assertEqual(eitem["loc_per_function"], 26.67)
        self.assertIn('commit_parents', eitem)
        self.assertIn('files_at_commit', eitem)

        item = self.items[3]
        eitem = enrich_backend.get_rich_items(item)[0]
        self.assertEqual(eitem['ccn'], None)
        self.assertEqual(eitem['num_funs'], None)
        self.assertEqual(eitem['tokens'], None)
        self.assertEqual(eitem['loc'], None)
        self.assertEqual(eitem['ext'], None)
        self.assertEqual(eitem['blanks'], None)
        self.assertEqual(eitem['comments'], None)
        self.assertEqual(eitem['file_path'], "tests/data/analyzers/sample_code.py")
        self.assertEqual(eitem['modules'], ["tests", "tests/data", "tests/data/analyzers"])
        self.assertEqual(eitem["comments_per_loc"], None)
        self.assertEqual(eitem["blanks_per_loc"], None)
        self.assertEqual(eitem["loc_per_function"], None)
        self.assertIn('commit_parents', eitem)
        self.assertIn('files_at_commit', eitem)

    def test_cocom_analysis_study(self):
        """ Test that the cocom analysis study works correctly """

        study, ocean_backend, enrich_backend = self._test_study('enrich_cocom_analysis')

        with self.assertLogs(logger, level='INFO') as cm:

            if study.__name__ == "enrich_cocom_analysis":
                study(ocean_backend, enrich_backend)
                self.assertEqual(cm.output[0], 'INFO:grimoire_elk.enriched.cocom:[cocom] study enrich-cocom-analysis '
                                               'start')
                self.assertEqual(cm.output[-1], 'INFO:grimoire_elk.enriched.cocom:[cocom] study enrich-cocom-analysis '
                                                'End')

    def test_perceval_params(self):
        """Test the extraction of perceval params from an URL"""

        url = "https://github.com/grimoirelab/perceval"
        expected_params = [
            'https://github.com/grimoirelab/perceval'
        ]
        self.assertListEqual(GraalOcean.get_perceval_params_from_url(url), expected_params)

        url = "https://github.com/grimoirelab/perceval /tmp/perceval-repo"
        expected_params = [
            'https://github.com/grimoirelab/perceval'
        ]
        self.assertListEqual(GraalOcean.get_perceval_params_from_url(url), expected_params)

    def test_copy_raw_fields(self):
        """Test copied raw fields"""

        self._test_raw_to_enrich()
        enrich_backend = self.connectors[self.connector][2]()

        for item in self.items:
            eitem = enrich_backend.get_rich_items(item)[0]
            for attribute in enrich_backend.RAW_FIELDS_COPY:
                if attribute in item:
                    self.assertEqual(item[attribute], eitem[attribute])
                else:
                    self.assertIsNone(eitem[attribute])


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(message)s')
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("requests").setLevel(logging.WARNING)
    unittest.main(warnings='ignore')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Quan Zhou <quan@bitergia.com>
#


import logging

from grimoirelab_toolkit.datetime import str_to_datetime

from .enrich import Enrich, metadata
from ..elastic_mapping import Mapping as BaseMapping


logger = logging.getLogger(__name__)


class Mapping(BaseMapping):

    @staticmethod
    def get_elastic_mappings(es_major):
        """Get Elasticsearch mapping.

        :param es_major: major version of Elasticsearch, as string
        :returns: dictionary with a key, 'items', with the mapping
        """

        mapping = """
        {
            "properties": {
                "text_analyzed": {
                  "type": "text",
                  "fielddata": true,
                  "index": true
                }
           }
        } """

        return {"items": mapping}


class WeblateEnrich(Enrich):

    mapping = Mapping

    def __init__(self, db_sortinghat=None, json_projects_map=None,
                 db_user='', db_password='', db_host='', db_path=None,
                 db_port=None, db_ssl=False, db_verify_ssl=True, db_tenant=None):
        super().__init__(db_sortinghat=db_sortinghat, json_projects_map=json_projects_map,
                         db_user=db_user, db_password=db_password, db_host=db_host,
                         db_port=db_port, db_path=db_path, db_ssl=db_ssl, db_verify_ssl=db_verify_ssl,
                         db_tenant=db_tenant)

        self.studies = []
        self.studies.append(self.enrich_demography)

    def get_field_author(self):
        return "author_data"

    def get_sh_identity(self, item, identity_field=None):
        # email not available for gitter
        identity = {
            'username': None,
            'name': None,
            'email': None
        }

        author = item['data'][self.get_field_author()] \
            if self.get_field_author() in item['data'] else None
        if not author:
            return identity

        identity['username'] = author.get('username', None)
        identity['name'] = author.get('full_name', None)
        identity['email'] = author.get('email', None)

        return identity

    def get_identities(self, item):
        """ Return the identities from an item """

        identity = self.get_sh_identity(item)
        yield identity

    def get_project_repository(self, eitem):
        repo = eitem['origin']
        return repo

    @metadata
    def get_rich_item(self, item):

        eitem = {}

        self.copy_raw_fields(self.RAW_FIELDS_COPY, item, eitem)

        change = item['data']

        change_timestamp = str_to_datetime(eitem['metadata__updated_on'])
        eitem['tz'] = int(change_timestamp.strftime("%H"))
        eitem['action'] = change['action']
        eitem['action_name'] = change['action_name']
        eitem['glossary_term'] = change.get('glossary_term', None)
        eitem['change_id'] = change['id']
        eitem['target'] = change['target']
        eitem['change_api_url'] = change['url']

        eitem['author_api_url'] = change['author']
        author_data = change.get('author_data', None)
        if author_data:
            for data in author_data:
                eitem['author_' + data] = author_data[data]

        eitem['user_api_url'] = change.get('user', None)
        if eitem['user_api_url']:
            username = eitem['user_api_url']
            eitem['user_name'] = username.split('api/users/')[1].split('/')[0] if username else None

        user_data = change.get('user_data', None)
        if user_data:
            for data in user_data:
                eitem['user_' + data] = user_data[data]

        eitem['unit_api_url'] = change['unit']
        unit_data = change.get('unit_data', None)
        if unit_data:
            for data in unit_data:
                if 'has_' in data:
                    eitem['unit_' + data] = 1 if unit_data[data] else 0
                else:
                    eitem['unit_' + data] = unit_data[data]

        eitem['component_api_url'] = change.get('component', None)
        if eitem['component_api_url']:
            component = eitem['component_api_url']
            # component = <source>/api/components/<project>/<component>
            project_name = component.split('api/components/')[1].split('/')[0] if component else None
            component_name = component.split('api/components/')[1].split('/')[1] if component else None
            eitem['project_name'] = project_name
            eitem['component_name'] = component_name

        eitem['translation_api_url'] = change.get('translation', None)
        if eitem['translation_api_url']:
            translation = eitem['translation_api_url']
            # component = <source>/api/translations/<project>/<component>/<translation>
            project_name = translation.split('api/translations/')[1].split('/')[0] if translation else None
            component_name = translation.split('api/translations/')[1].split('/')[1] if translation else None
            translation_name = translation.split('api/translations/')[1].split('/')[2] if translation else None
            eitem['project_name'] = project_name
            eitem['component_name'] = component_name
            eitem['translation_name'] = translation_name

        if self.sortinghat:
            eitem.update(self.get_item_sh(item))

        if self.prjs_map:
            eitem.update(self.get_item_project(eitem))

        eitem.update(self.get_grimoire_fields(item["metadata__updated_on"], "message"))

        self.add_repository_labels(eitem)
        self.add_metadata_filter_raw(eitem)
        return eitem

    def enrich_demography(self, ocean_backend, enrich_backend, alias, date_field="grimoire_creation_date",
                          author_field="author_uuid"):

        super().enrich_demography(ocean_backend, enrich_backend, alias, date_field, author_field=author_field)
